# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.
# pylint: disable-msg=E1101

from typing import Any
from collections.abc import Sequence

from xpra.util.env import envint
from xpra.os_util import OSX
from xpra.net.common import PacketType
from xpra.util.version import vtrim
from xpra.scripts.config import parse_bool_or_int
from xpra.codecs.constants import preforder, STREAM_ENCODINGS, TRUE_LOSSLESS_ENCODINGS
from xpra.codecs.loader import get_codec, codec_versions, load_codec, unload_codecs
from xpra.codecs.video import getVideoHelper
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger
from xpra.common import FULL_INFO

log = Logger("encoding")

INIT_DELAY = envint("XPRA_ENCODER_INIT_DELAY", 0)

ENCODINGS_WITH_SPEED = (
    "h264", "vp8", "vp9",
    "rgb24", "rgb32",
    "png", "png/P", "png/L", "webp",
    "scroll",
)
ENCODINGS_WITH_QUALITY = (
    "jpeg", "webp",
    "h264", "vp8", "vp9",
    "scroll",
)

ENCODING_OPTIONS = (
    "quality", "min-quality", "max-quality",
    "speed", "min-speed", "max-speed",
)


class EncodingServer(StubServerMixin):
    """
    Mixin for adding encodings to a server
    """
    PREFIX = "encoding"

    def __init__(self):
        self.default_quality = -1
        self.default_min_quality = 0
        self.default_speed = -1
        self.default_min_speed = 0
        self.encoding = ""
        self.allowed_encodings: Sequence[str] = ()
        self.core_encodings: Sequence[str] = ()
        self.encodings: Sequence[str] = ()
        self.lossless_encodings: Sequence[str] = ()
        self.lossless_mode_encodings: Sequence[str] = ()
        self.default_encoding: str = ""
        self.scaling_control = None
        self.video = True

    def init(self, opts) -> None:
        self.encoding = opts.encoding
        self.allowed_encodings = opts.encodings
        self.default_quality = opts.quality
        self.default_min_quality = opts.min_quality
        self.default_speed = opts.speed
        self.default_min_speed = opts.min_speed
        self.video = opts.video
        if self.video:
            if opts.video_scaling.lower() not in ("auto", "on"):
                self.scaling_control = parse_bool_or_int("video-scaling", opts.video_scaling)
            getVideoHelper().set_modules(video_encoders=opts.video_encoders, csc_modules=opts.csc_modules)

    def setup(self) -> None:
        # essential codecs, load them early:
        load_codec("enc_rgb")
        load_codec("enc_pillow")
        ae = self.allowed_encodings
        if "webp" in ae:
            # try to load the fast webp encoder:
            load_codec("enc_webp")
        if "png" in ae or "png/L" in ae:
            # try to load the fast png encoder:
            load_codec("enc_spng")
        if "jpeg" in ae or "jpega" in ae:
            # try to load the fast jpeg encoders:
            load_codec("enc_jpeg")
        if "avif" in ae:
            load_codec("enc_avif")
        self.init_encodings()
        self.add_init_thread_callback(self.reinit_encodings)

    def reinit_encodings(self) -> None:
        self.init_encodings()
        # any window mapped before the threaded init completed
        # may need to re-initialize its list of encodings:
        log("reinit_encodings()")
        try:
            from xpra.server.source.windows import WindowsMixin
        except ImportError:
            return
        for ss in self._server_sources.values():
            if isinstance(ss, WindowsMixin):
                ss.reinit_encodings(self)

    def threaded_setup(self) -> None:
        if INIT_DELAY > 0:
            from time import sleep
            sleep(INIT_DELAY)
        # load the slower codecs
        if "jpeg" in self.allowed_encodings and not OSX:
            load_codec("enc_nvjpeg")
        if self.video:
            # load video codecs:
            getVideoHelper().init()
        self.init_encodings()

    def cleanup(self) -> None:
        getVideoHelper().cleanup()
        unload_codecs()

    def get_server_features(self, _source=None) -> dict[str, Any]:
        return {}

    def get_info(self, _proto) -> dict[str, Any]:
        info = {
            "encodings": self.get_encoding_info(),
        }
        if self.video:
            info["video"] = getVideoHelper().get_info()
        if FULL_INFO > 0:
            for k, v in codec_versions.items():
                info.setdefault("encoding", {}).setdefault(k, {})["version"] = vtrim(v)
        return info

    def get_encoding_info(self) -> dict[str, Any]:
        info = {
            "": self.encodings,  # redundant since v6
            "core": self.core_encodings,
            "allowed": self.allowed_encodings,
            "lossless": self.lossless_encodings,
            "with_speed": tuple(set({"rgb32": "rgb", "rgb24": "rgb"}.get(x, x)
                                    for x in self.core_encodings if x in ENCODINGS_WITH_SPEED)),
            "with_quality": tuple(x for x in self.core_encodings if x in ENCODINGS_WITH_QUALITY),
            "with_lossless_mode": self.lossless_mode_encodings,
        }
        info.update(self.get_encoding_settings())
        return info

    def get_encoding_settings(self) -> dict[str, Any]:
        info = {}
        for prop, value in {
            "quality": self.default_quality,
            "min-quality": self.default_min_quality,
            "speed": self.default_speed,
            "min-speed": self.default_min_speed,
        }.items():
            if value > 0:
                info[prop] = value
        return info

    def init_encodings(self) -> None:
        encs: list[str] = []
        core_encs: list[str] = []
        lossless: list[str] = []
        log("init_encodings() allowed_encodings=%s", self.allowed_encodings)

        def add_encoding(encoding: str) -> None:
            log("add_encoding(%s)", encoding)
            enc = {"rgb32": "rgb", "rgb24": "rgb"}.get(encoding, encoding)
            if self.allowed_encodings is not None:
                if enc not in self.allowed_encodings and encoding not in self.allowed_encodings:
                    # not in whitelist (if it exists)
                    return
            if enc not in encs:
                encs.append(enc)
            if encoding not in core_encs:
                core_encs.append(encoding)
            if encoding in TRUE_LOSSLESS_ENCODINGS and encoding not in lossless:
                lossless.append(encoding)

        def add_encodings(*encodings: str) -> None:
            log("add_encodings%s", encodings)
            for enc in encodings:
                add_encoding(enc)

        add_encodings("rgb24", "rgb32")
        try:
            from xpra.server.window.motion import ScrollData  # @UnresolvedImport
            assert ScrollData
            add_encoding("scroll")
        except (ImportError, TypeError) as e:
            log.error("Error: 'scroll' encoding is not available")
            log.estr(e)
        add_encoding("scroll")

        # video encoders (empty when first called - see threaded_init)
        ve = getVideoHelper().get_encodings()
        log("init_encodings() adding video encodings: %s", ve)
        add_encodings(*ve)  # ie: ["vp8", "h264"]
        # Pithon Imaging Library:
        enc_pillow = get_codec("enc_pillow")
        log("enc_pillow=%s", enc_pillow)
        if enc_pillow:
            pil_encs = enc_pillow.get_encodings()
            log("pillow encodings: %s", pil_encs)
            for encoding in pil_encs:
                add_encoding(encoding)
        for codec_name in ("enc_avif", "enc_jpeg", "enc_nvjpeg"):
            codec = get_codec(codec_name)
            if codec:
                add_encodings(*codec.get_encodings())
        # look for video encodings with lossless mode:
        for enc in ve:
            for colorspace, especs in getVideoHelper().get_encoder_specs(enc).items():
                for espec in especs:
                    if espec.has_lossless_mode and enc not in lossless:
                        log("found lossless mode for encoding %s with %s and colorspace %s", enc, espec, colorspace)
                        lossless.append(enc)
                        break
        # now update the variables:
        encs.append("grayscale")
        if any(enc in encs for enc in STREAM_ENCODINGS):
            encs.append("stream")
        self.encodings = preforder(encs)
        self.core_encodings = preforder(core_encs)
        self.lossless_mode_encodings = preforder(lossless)
        self.lossless_encodings = preforder(enc for enc in core_encs
                                            if (enc.startswith("png") or enc.startswith("rgb") or enc == "webp"))
        log("allowed encodings=%s, encodings=%s, core encodings=%s, lossless encodings=%s",
            self.allowed_encodings, encs, core_encs, self.lossless_encodings)
        self.default_encoding = self.encodings[0]
        # default encoding:
        if not self.encoding or str(self.encoding).lower() in ("auto", "none"):
            self.default_encoding = ""
        elif self.encoding in self.encodings:
            log.warn("ignored invalid default encoding option: %s", self.encoding)
            self.default_encoding = self.encoding

    def _process_encoding_set(self, proto, packet: PacketType) -> None:
        encoding = str(packet[1])
        ss = self.get_server_source(proto)
        if ss is None:
            return
        if len(packet) >= 3:
            # client specified which windows this is for:
            in_wids = packet[2]
            wids = []
            wid_windows = {}
            for wid in in_wids:
                if wid not in self._id_to_window:
                    continue
                wids.append(wid)
                wid_windows[wid] = self._id_to_window.get(wid)
        else:
            # apply to all windows:
            wids = None
            wid_windows = self._id_to_window
        ss.set_encoding(encoding, wids)
        self._refresh_windows(proto, wid_windows, {})

    def _process_quality(self, proto, packet) -> None:
        self._modify_sq(proto, "quality", packet[1])

    def _process_min_quality(self, proto, packet) -> None:
        self._modify_sq(proto, "min-quality", packet[1])

    def _process_max_quality(self, proto, packet) -> None:
        self._modify_sq(proto, "max-quality", packet[1])

    def _process_speed(self, proto, packet) -> None:
        self._modify_sq(proto, "speed", packet[1])

    def _process_min_speed(self, proto, packet) -> None:
        self._modify_sq(proto, "min-speed", packet[1])

    def _process_max_speed(self, proto, packet) -> None:
        self._modify_sq(proto, "max-speed", packet[1])

    def _modify_sq(self, proto, attr: str, value: int) -> None:
        """ modify speed or quality attributes """
        ss = self.get_server_source(proto)
        if not ss:
            return
        assert attr in ENCODING_OPTIONS
        log("Setting %s to %s", attr, value)
        fn = getattr(ss, "set_%s" % attr.replace("-", "_"))
        fn(value)
        self.call_idle_refresh_all_windows(proto)

    def call_idle_refresh_all_windows(self, proto) -> None:
        # we can't assume that the window server mixin is loaded:
        refresh = getattr(self, "_idle_refresh_all_windows", None)
        if refresh:
            refresh(proto)  # pylint: disable=not-callable

    def _process_encoding_options(self, proto, packet) -> None:
        ss = self.get_server_source(proto)
        if not ss:
            return
        updates = packet[1]
        for attr, value in updates.items():
            if attr not in ENCODING_OPTIONS:
                log.warn(f"Warning: {attr!r} is not a valid encoding option")
                continue
            log(f"setting {attr!r} to {value!r}")
            fn = getattr(ss, "set_%s" % attr.replace("-", "_"))
            fn(value)
        self.call_idle_refresh_all_windows(proto)

    def init_packet_handlers(self) -> None:
        self.add_packets(f"{EncodingServer.PREFIX}-set", f"{EncodingServer.PREFIX}-options")
        # legacy:
        self.add_packets(
            "quality", "min-quality", "max-quality",
            "speed", "min-speed", "max-speed",
        )
        self.add_legacy_alias("encodings", f"{EncodingServer.PREFIX}-set")



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
import socket
from collections.abc import Sequence

from xpra.util.objects import typedict
from xpra.util.str_fn import obsc, strtobytes
from xpra.util.env import envint
from xpra.util.io import stderr_print
from xpra.auth.sys_auth_base import SysAuthenticatorBase, log
from xpra.auth.common import parse_uid, parse_gid
from xpra.log import is_debug_enabled, consume_verbose_argv

LDAP_REFERRALS = envint("XPRA_LDAP_REFERRALS", 0)
LDAP_PROTOCOL_VERSION = envint("XPRA_LDAP_PROTOCOL_VERSION", 3)
LDAP_TRACE_LEVEL = envint("XPRA_LDAP_TRACE_LEVEL")
LDAP_CACERTFILE = os.environ.get("XPRA_LDAP_CACERTFILE")
LDAP_ENCODING = os.environ.get("XPRA_LDAP_ENCODING", "utf-8")
LDAP_USERNAME_FORMAT = os.environ.get("XPRA_LDAP_USERNAME_FORMAT", "cn=%username, o=%domain")


class Authenticator(SysAuthenticatorBase):
    CLIENT_USERNAME = True

    def __init__(self, **kwargs):
        self.tls = bool(int(kwargs.pop("tls", "0")))
        self.host = kwargs.pop("host", "localhost")
        self.cacert = kwargs.pop("cacert", LDAP_CACERTFILE)
        self.encoding = kwargs.pop("encoding", LDAP_ENCODING)
        self.uid = parse_uid(kwargs.pop("uid", None))
        self.gid = parse_gid(kwargs.pop("gid", None))
        if self.tls:
            default_port = 636
        else:
            default_port = 389
        self.port = int(kwargs.pop("port", default_port))
        self.username_format = kwargs.pop("username_format", LDAP_USERNAME_FORMAT)
        # self.username_format = kwargs.pop("username_format", "%username@%domain")
        super().__init__(**kwargs)
        log("ldap auth: host=%s, port=%i, tls=%s, username_format=%s, cacert=%s, encoding=%s",
            self.host, self.port, self.tls, self.username_format, self.cacert, self.encoding)

    def get_uid(self) -> int:
        return self.uid

    def get_gid(self) -> int:
        return self.gid

    def __repr__(self):
        return "ldap"

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        if "xor" not in digests:
            log.error("Error: ldap authentication requires the 'xor' digest")
            return b"", ""
        return super().get_challenge(("xor", ))

    def check_password(self, password: str) -> bool:
        log("check_password(%s)", obsc(password))

        def emsg(e):
            try:
                log.warn(" LDAP Error: %s", e.message["desc"])
                if "info" in e.message:
                    log.warn("  %s", e.message["info"])
            except Exception:
                # python3: no way to get to the message dict?
                log.warn(" %s", e)

        try:
            from ldap import (  # pylint: disable=import-outside-toplevel
                initialize, LDAPError,
                INVALID_CREDENTIALS, SERVER_DOWN,
                OPT_REFERRALS,
                OPT_X_TLS_CACERTFILE,
            )
        except ImportError as e:
            log("check(..)", exc_info=True)
            log.warn("Warning: cannot use ldap authentication:")
            log.warn(" %s", e)
            return False
        try:
            assert self.username and password
            if self.tls:
                protocol = "ldaps"
            else:
                protocol = "ldap"
            server = "%s://%s:%i" % (protocol, self.host, self.port)
            conn = initialize(server, trace_level=LDAP_TRACE_LEVEL or is_debug_enabled("auth"))
            conn.protocol_version = LDAP_PROTOCOL_VERSION
            conn.set_option(OPT_REFERRALS, LDAP_REFERRALS)
            if self.cacert:
                conn.set_option(OPT_X_TLS_CACERTFILE, self.cacert)
            log("ldap.open(%s)=%s", server, conn)
            try:
                domain = socket.getfqdn().split(".", 1)[1]
            except Exception:
                domain = "localdomain"
            user = self.username_format.replace("%username", self.username).replace("%domain", domain)
            log("user=%s", user)
            try:
                pvalue = password.encode(self.encoding)
                log(f"ldap encoded password as {self.encoding}")
            except Exception:
                pvalue = strtobytes(password)
            conn.simple_bind_s(user, pvalue)
            log("simple_bind_s(%s, %s) done", user, obsc(password))
            return True
        except INVALID_CREDENTIALS:
            log("check(..)", exc_info=True)
        except SERVER_DOWN as e:
            log("check(..)", exc_info=True)
            log.warn("Warning: LDAP %sserver at %s:%i is unreachable", ["", "TLS "][self.tls], self.host, self.port)
            emsg(e)
        except LDAPError as e:
            log("check(..)", exc_info=True)
            log.warn("Error: ldap authentication failed:")
            emsg(e)
        return False


def main(argv) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.net.digest import get_salt, get_digests, gendigest
    from xpra.platform import program_context
    with program_context("LDAP-Password-Auth", "LDAP-Password-Authentication"):
        consume_verbose_argv(argv, "auth")
        if len(argv) not in (3, 4, 5, 6, 7):
            stderr_print("%s invalid arguments" % argv[0])
            stderr_print("usage: %s username password [host] [port] [tls] [username_format]" % argv[0])
            return 1
        username = argv[1]
        password = argv[2]
        kwargs = {"username": username}
        if len(argv) >= 4:
            kwargs["host"] = argv[3]
        if len(argv) >= 5:
            kwargs["port"] = argv[4]
        if len(argv) >= 6:
            kwargs["tls"] = argv[5]
        if len(argv) >= 7:
            kwargs["username_format"] = argv[6]
        a = Authenticator(**kwargs)
        server_salt, digest = a.get_challenge(("xor", ))
        salt_digest = a.choose_salt_digest(get_digests())
        assert digest == "xor"
        client_salt = get_salt(len(server_salt))
        combined_salt = gendigest(salt_digest, client_salt, server_salt)
        assert digest == "xor"
        response = gendigest(digest, password, combined_salt)
        caps = typedict({
            "challenge_response": response,
            "challenge_client_salt": client_salt,
        })
        r = a.authenticate(caps)
        print("success: %s" % bool(r))
        return int(not r)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic

from xpra.os_util import gi_import
from xpra.client.gui.tray_base import TrayBase
from xpra.gtk.pixbuf import get_pixbuf_from_data
from xpra.platform.darwin.menu import getOSXMenuHelper
from xpra.platform.darwin import set_exit_cb
from xpra.platform.gui import ready as gui_ready
from xpra.log import Logger

GdkPixbuf = gi_import("GdkPixbuf")

log = Logger("tray", "osx")

# constants for attention_request:
CRITICAL_REQUEST = 0
INFO_REQUEST = 10


class OSXTray(TrayBase):

    def __init__(self, *args):
        super().__init__(*args)
        from xpra.platform.darwin import get_OSXApplication
        self.macapp = get_OSXApplication()
        assert self.macapp, "cannot use OSX Tray without the native gtkosx_application bindings"
        self.last_attention_request_id = -1

        self.set_global_menu()
        self.set_dock_menu()
        self.set_dock_icon()
        set_exit_cb(self.quit)

    def get_geometry(self):
        return None

    def show(self) -> None:
        """
        This cannot be implemented on MacOS,
        as the dock icon is always shown
        """

    def hide(self) -> None:
        """ Unfortunately, the dock icon cannot be hidden """

    def quit(self, *args) -> bool:
        log("quit(%s) exit_cb=%s", args, self.exit_cb)
        if self.exit_cb:
            self.exit_cb()
            return True  # we've handled the quit request ourselves - I hope..
        return False

    def ready(self) -> None:
        gui_ready()

    def set_tooltip(self, tooltip: str = "") -> None:
        # label cannot be set on the dock icon?
        pass

    def set_blinking(self, on: bool) -> None:
        if on:
            if self.last_attention_request_id < 0:
                self.last_attention_request_id = self.macapp.attention_request(INFO_REQUEST)
        else:
            if self.last_attention_request_id >= 0:
                self.macapp.cancel_attention_request(self.last_attention_request_id)
                self.last_attention_request_id = -1

    def set_icon_from_data(self, pixels, has_alpha: bool, w: int, h: int, rowstride: int, options=None) -> None:
        tray_icon = get_pixbuf_from_data(pixels, has_alpha, w, h, rowstride)
        self.macapp.set_dock_icon_pixbuf(tray_icon)
        self.icon_timestamp = monotonic()

    def do_set_icon_from_file(self, filename: str) -> None:
        if not self.macapp:
            return
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename=filename)
        self.macapp.set_dock_icon_pixbuf(pixbuf)
        self.icon_timestamp = monotonic()

    def set_global_menu(self) -> None:
        mh = getOSXMenuHelper()
        if mh.build() != self.menu:
            log.error("the menu (%s) is not from the menu helper!", self.menu)
            return
        # redundant: the menu bar has already been set during gui init
        # using the basic the simple menu from build_menu_bar()
        import warnings
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message=".*invalid cast from 'GtkMenuBar'")
            self.macapp.set_menu_bar(self.menu)
        mh.add_full_menu()
        log("OSXTray.set_global_menu() done")

    def set_dock_menu(self) -> None:
        # dock menu
        log("OSXTray.set_dock_menu()")
        Gtk = gi_import("Gtk")
        self.dock_menu = Gtk.Menu()
        self.disconnect_dock_item = Gtk.MenuItem(label="Disconnect")
        self.disconnect_dock_item.connect("activate", self.quit)
        self.dock_menu.add(self.disconnect_dock_item)
        self.dock_menu.show_all()
        self.macapp.set_dock_menu(self.dock_menu)
        log("OSXTray.set_dock_menu() done")

    def set_dock_icon(self) -> None:
        filename = self.get_icon_filename()
        if not filename:
            log.warn("Warning: cannot set dock icon, file not found!")
            return
        log("OSXTray.set_dock_icon() loading icon from %s", filename)
        pixbuf = GdkPixbuf.Pixbuf.new_from_file(filename=filename)
        self.macapp.set_dock_icon_pixbuf(pixbuf)



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from typing import Any
from collections.abc import Callable, Sequence

from xpra.client.gui.factory import get_client_base_classes
from xpra.client.base.client import XpraClientBase
from xpra.client.gui.keyboard_helper import KeyboardHelper
from xpra.platform import set_name
from xpra.platform.gui import ready as gui_ready, get_wm_name, get_session_type, ClientExtras
from xpra.common import FULL_INFO, noop, NotificationID, ConnectionMessage, noerr, get_run_info
from xpra.net.common import PacketType
from xpra.keyboard.common import KeyEvent
from xpra.os_util import POSIX, WIN32, OSX, gi_import
from xpra.util.child_reaper import reaper_cleanup
from xpra.util.version import get_platform_info
from xpra.util.system import is_Wayland, platform_name
from xpra.util.objects import typedict
from xpra.util.screen import log_screen_sizes
from xpra.util.str_fn import std, Ellipsizer, repr_ellipsized
from xpra.util.env import envint, envbool
from xpra.scripts.config import str_to_bool
from xpra.exit_codes import ExitCode, ExitValue
from xpra.client.base import features
from xpra.log import Logger

CLIENT_BASES = get_client_base_classes()
ClientBaseClass = type('ClientBaseClass', CLIENT_BASES, {})

GLib = gi_import("GLib")

log = Logger("client")
keylog = Logger("client", "keyboard")
log("UIXpraClient%s: %s", ClientBaseClass, CLIENT_BASES)

NOTIFICATION_EXIT_DELAY = envint("XPRA_NOTIFICATION_EXIT_DELAY", 2)
MOUSE_DELAY_AUTO = envbool("XPRA_MOUSE_DELAY_AUTO", True)
DELAY_KEYBOARD_DATA = envbool("XPRA_DELAY_KEYBOARD_DATA", True)


class UIXpraClient(ClientBaseClass):
    """
    Utility superclass for client classes which have a UI.
    See gtk_client_base and its subclasses.
    """
    # NOTE: these signals aren't registered here because this class
    # does not extend GObject,
    # the gtk client subclasses will take care of it.
    # these are all "no-arg" signals
    __signals__ = ["first-ui-received", ]
    for c in CLIENT_BASES:
        if c != XpraClientBase:
            __signals__ += c.__signals__

    # noinspection PyMissingConstructor
    def __init__(self):  # pylint: disable=super-init-not-called
        # try to ensure we start on a new line (see #4023):
        noerr(sys.stdout.write, "\n")
        run_info = get_run_info(f"{self.client_toolkit()} client")
        wm = get_wm_name()  # pylint: disable=assignment-from-none
        if wm:
            run_info.append(f" window manager is {wm!r}")
        for info in run_info:
            log.info(info)
        # same for tray:
        self.tray = None
        for c in CLIENT_BASES:
            log("calling %s.__init__()", c)
            c.__init__(self)  # pylint: disable=non-parent-init-called
        self._ui_events: int = 0
        self.title: str = ""
        self.session_name: str = ""

        self.server_platform: str = ""
        self.server_session_name: str = ""

        # features:
        self.opengl_enabled: bool = False
        self.opengl_props: dict[str, Any] = {}
        self.readonly: bool = False
        self.xsettings_enabled: bool = False
        self.server_start_new_commands: bool = False
        self.server_xdg_menu = None
        self.start_new_commands: bool = False
        self.request_start = []
        self.request_start_child = []
        self.headerbar = None
        self.suspended = False

        # in WindowClient - should it be?
        # self.server_is_desktop = False
        self.server_sharing: bool = False
        self.server_sharing_toggle: bool = False
        self.server_lock: bool = False
        self.server_lock_toggle: bool = False
        self.server_keyboard: bool = True
        self.server_pointer: bool = True
        self.server_commands_info = False
        self.server_commands_signals: Sequence[str] = ()
        self.server_readonly = False
        self.server_setting_updates: dict[str, Any] = {}

        self.client_supports_opengl: bool = False
        self.client_supports_sharing: bool = False
        self.client_lock: bool = False

        # helpers and associated flags:
        self.client_extras = None
        self._mouse_position_delay = 5
        self.keyboard_helper_class: type = KeyboardHelper
        self.keyboard_helper = None
        self.keyboard_grabbed: bool = False
        self.keyboard_sync: bool = False
        self.key_repeat_delay = -1
        self.key_repeat_interval = -1
        self.kh_warning: bool = False
        self.menu_helper = None

        # state:
        self._on_handshake: Sequence[tuple[Callable, Sequence[Any]]] | None = []
        self._on_server_setting_changed: dict[str, Sequence[Callable[[str, Any], None]]] = {}

    def init(self, opts) -> None:
        """ initialize variables from configuration """
        for c in CLIENT_BASES:
            log(f"init: {c}")
            c.init(self, opts)

        self.title = opts.title
        self.session_name = opts.session_name
        self.xsettings_enabled = not (OSX or WIN32) and str_to_bool(opts.xsettings)
        self.readonly = opts.readonly
        self.client_supports_sharing = opts.sharing is True
        self.client_lock = opts.lock is True
        self.headerbar = opts.headerbar

    def client_toolkit(self) -> str:
        raise NotImplementedError()

    def init_ui(self, opts) -> None:
        """ initialize user interface """

        def noauto(val) -> str | Sequence | None:
            default = [] if isinstance(val, Sequence) else None
            if not val:
                return default
            if str(val).lower() == "auto":
                return default
            return val

        send_keyboard = noop
        if not self.readonly:
            def do_send_keyboard(*parts):
                self.after_handshake(self.send, *parts)

            send_keyboard = do_send_keyboard
        try:
            kwargs = dict(
                (x, noauto(getattr(opts, "keyboard_%s" % x))) for x in (
                    "backend", "model", "layout", "layouts", "variant", "variants", "options",
                )
            )
            self.keyboard_helper = self.keyboard_helper_class(send_keyboard, opts.keyboard_sync,
                                                              opts.shortcut_modifiers,
                                                              opts.key_shortcut,
                                                              opts.keyboard_raw, **kwargs)
            if DELAY_KEYBOARD_DATA and not self.readonly:
                self.after_handshake(self.keyboard_helper.send_keymap)
        except ImportError as e:
            keylog("error instantiating %s", self.keyboard_helper_class, exc_info=True)
            keylog.warn(f"Warning: no keyboard support, {e}")

        if features.windows:
            self.init_opengl(opts.opengl)

        if ClientExtras is not None:
            self.client_extras = ClientExtras(self, opts)  # pylint: disable=not-callable

        self.start_new_commands = str_to_bool(opts.start_new_commands)
        if self.start_new_commands and (opts.start or opts.start_child):
            from xpra.scripts.main import strip_defaults_start_child
            from xpra.scripts.config import make_defaults_struct
            defaults = make_defaults_struct()
            self.request_start = strip_defaults_start_child(opts.start, defaults.start)
            self.request_start_child = strip_defaults_start_child(opts.start_child, defaults.start_child)

        if MOUSE_DELAY_AUTO:
            try:
                # some platforms don't detect the vrefresh correctly
                # (ie: macos in virtualbox?), so use a sane default minimum
                # discount by 5ms to ensure we have time to hit the target
                v = max(60, self.get_vrefresh())
                self._mouse_position_delay = max(5, 1000 // v // 2 - 5)
                log(f"mouse position delay: {self._mouse_position_delay}")
            except (AttributeError, OSError):
                log("failed to calculate automatic delay", exc_info=True)

    def get_vrefresh(self) -> int:
        # this method is overridden in the GTK client
        from xpra.platform.gui import get_vrefresh  # pylint: disable=import-outside-toplevel
        return get_vrefresh()

    def run(self) -> ExitValue:
        if self.client_extras:
            GLib.idle_add(self.client_extras.ready)
        for c in CLIENT_BASES:
            c.run(self)
        return self.exit_code or 0

    def quit(self, exit_code: ExitValue = 0) -> None:
        raise NotImplementedError()

    def cleanup(self) -> None:
        log("UIXpraClient.cleanup()")
        for c in CLIENT_BASES:
            c.cleanup(self)
        for x in (self.keyboard_helper, self.tray, self.menu_helper, self.client_extras):
            if x is None:
                continue
            log(f"UIXpraClient.cleanup() calling {type(x)}.cleanup()")
            with log.trap_error(f"Error on {type(x)} cleanup"):
                x.cleanup()
        # the protocol has been closed, it is now safe to close all the windows:
        # (cleaner and needed when we run embedded in the client launcher)
        reaper_cleanup()
        log("UIXpraClient.cleanup() done")

    def signal_cleanup(self) -> None:
        log("UIXpraClient.signal_cleanup()")
        XpraClientBase.signal_cleanup(self)
        reaper_cleanup()
        log("UIXpraClient.signal_cleanup() done")

    def get_info(self) -> dict[str, Any]:
        info: dict[str, Any] = {}
        if FULL_INFO > 0:
            info["session-name"] = self.session_name
        for c in CLIENT_BASES:
            with log.trap_error("Error collection information from %s", c):
                info.update(c.get_info(self))
        return info

    def suspend(self, *args) -> None:
        log("suspend%s", args)
        self.suspended = True
        for c in CLIENT_BASES:
            c.suspend(self)
        # tell the server:
        # ("ui" and "window-ids" arguments are optional since v6.3)
        self.send("suspend", True, tuple(self._id_to_window.keys()))

    def resume(self, *args) -> None:
        log("resume%s", args)
        self.suspended = False
        for c in CLIENT_BASES:
            c.resume(self)
        # tell the server:
        # ("ui" and "window-ids" arguments are optional since v6.3)
        self.send("resume", True, tuple(self._id_to_window.keys()))

    def show_about(self, *_args) -> None:
        log.warn(f"show_about() is not implemented in {self!r}")

    def show_session_info(self, *_args) -> None:
        log.warn(f"show_session_info() is not implemented in {self!r}")

    def show_bug_report(self, *_args) -> None:
        log.warn(f"show_bug_report() is not implemented in {self!r}")

    def init_opengl(self, _enable_opengl) -> None:
        self.opengl_enabled = False
        self.client_supports_opengl = False
        self.opengl_props = {"info": "not supported"}

    def _ui_event(self) -> None:
        if self._ui_events == 0:
            self.emit("first-ui-received")
        self._ui_events += 1

    def get_mouse_position(self):
        raise NotImplementedError()

    def get_current_modifiers(self):
        raise NotImplementedError()

    def send_start_new_commands(self) -> None:
        log(f"send_start_new_commands() {self.request_start=}, {self.request_start_child=}")
        import shlex
        for cmd in self.request_start:
            cmd_parts = shlex.split(cmd)
            self.send_start_command(cmd_parts[0], cmd_parts, True)
        for cmd in self.request_start_child:
            cmd_parts = shlex.split(cmd)
            self.send_start_command(cmd_parts[0], cmd_parts, False)

    def send_start_command(self, name: str, command: list[str], ignore: bool, sharing: bool = True) -> None:
        log("send_start_command%s", (name, command, ignore, sharing))
        assert name is not None and command is not None and ignore is not None
        self.send("start-command", name, command, ignore, sharing)

    ######################################################################
    # trigger notifications on disconnection,
    # and wait before actually exiting so the notification has a chance of being seen
    def server_disconnect_warning(self, reason, *info):
        if self.exit_code is None:
            body = "\n".join(info)
            if not self.connection_established:
                if ConnectionMessage.AUTHENTICATION_FAILED.value in info:
                    title = "Authentication failed"
                    self.exit_code = ExitCode.AUTHENTICATION_FAILED
                else:
                    title = "Connection failed: %s" % reason
                    self.exit_code = ExitCode.CONNECTION_FAILED
            else:
                if self.completed_startup:
                    title = "Xpra Session Disconnected: %s" % reason
                    self.exit_code = ExitCode.CONNECTION_LOST
                else:
                    title = "Connection failed during startup: %s" % reason
                    self.exit_code = ExitCode.CONNECTION_FAILED
            self.may_notify(NotificationID.DISCONNECT, title, body, icon_name="disconnected")
            # show text notification then quit:
            delay = NOTIFICATION_EXIT_DELAY * int(features.notifications)
            GLib.timeout_add(delay * 1000, XpraClientBase.server_disconnect_warning, self, title, *info)
        self.cleanup()

    def server_disconnect(self, reason: str, *info) -> None:
        body = "\n".join(info)
        self.may_notify(NotificationID.DISCONNECT,
                        f"Xpra Session Disconnected: {reason}", body, icon_name="disconnected")
        delay = NOTIFICATION_EXIT_DELAY * int(features.notifications)
        if self.exit_code is None:
            self.exit_code = self.server_disconnect_exit_code(reason, *info)
        GLib.timeout_add(delay * 1000, XpraClientBase.server_disconnect, self, reason, *info)
        self.cleanup()

    ######################################################################
    # hello:
    def make_hello(self) -> dict[str, Any]:
        caps = XpraClientBase.make_hello(self)
        # don't try to find the server uuid if this platform cannot run servers..
        # (doing so causes lockups on win32 and startup errors on osx)
        if POSIX and not (OSX or is_Wayland()):
            # we may be running inside another server!
            try:
                from xpra.x11.server.server_uuid import get_uuid, get_mode  # pylint: disable=import-outside-toplevel
                if get_mode() != "shadow":
                    uuid = get_uuid()
                    if uuid:
                        caps["server_uuid"] = uuid
            except (ImportError, RuntimeError):
                log("skipped server uuid lookup", exc_info=True)
        for x in (  # generic feature flags:
                "xdg-menu-update",
                "setting-change", "mouse",
        ):
            caps[x] = True
        caps.setdefault("wants", []).append("events")
        caps |= {
            # generic server flags:
            "share": self.client_supports_sharing,
            "lock": self.client_lock,
            "xdg-menu": self.start_new_commands,
        }
        caps.update(self.get_keyboard_caps())
        for c in CLIENT_BASES:
            caps.update(c.get_caps(self))
        if FULL_INFO > 0:
            def skipkeys(d, *keys):
                return {k: v for k, v in d.items() if k not in keys}

            pi = get_platform_info()
            op = self.opengl_props
            if FULL_INFO == 1:
                op = skipkeys(op, "extensions", "GLU.extensions")
                pi = skipkeys(pi, "release", "sysrelease", "platform", "processor", "architecture")
            caps["platform"] = pi
            caps["opengl"] = op
            caps["session-type"] = get_session_type()
        return caps

    ######################################################################
    # connection setup:
    def setup_connection(self, conn) -> None:
        super().setup_connection(conn)
        for c in CLIENT_BASES:
            if c != XpraClientBase:
                c.setup_connection(self, conn)

    def server_connection_established(self, caps: typedict) -> bool:
        if not XpraClientBase.server_connection_established(self, caps):
            return False
        # process the rest from the UI thread:
        GLib.idle_add(self.process_ui_capabilities, caps)
        self.add_control_commands()
        return True

    def parse_server_capabilities(self, c: typedict) -> bool:
        for cb in CLIENT_BASES:
            log("%s.parse_server_capabilities(..)", cb)
            try:
                if not cb.parse_server_capabilities(self, c):
                    log.info(f"failed to parse server capabilities in {cb}")
                    return False
            except Exception as e:
                log("%s.parse_server_capabilities(%s)", cb, Ellipsizer(c), exc_info=True)
                log.error("Error parsing server capabilities using %s: %s", cb, e)
                return False
        self.server_session_name = c.strget("session_name")
        set_name("Xpra", self.session_name or self.server_session_name or "Xpra")
        self.server_platform = c.strget("platform")
        self.server_sharing = c.boolget("sharing")
        self.server_sharing_toggle = c.boolget("sharing-toggle")
        self.server_lock = c.boolget("lock")
        self.server_lock_toggle = c.boolget("lock-toggle")
        self.server_keyboard = c.boolget("keyboard", True)
        self.server_pointer = c.boolget("pointer", True)
        self.server_start_new_commands = c.boolget("start-new-commands")
        if self.server_start_new_commands:
            self.server_xdg_menu = c.dictget("xdg-menu", None)
        if self.request_start or self.request_start_child:
            if self.server_start_new_commands:
                self.after_handshake(self.send_start_new_commands)
            else:
                log.warn("Warning: cannot start new commands")
                log.warn(" the feature is currently disabled on the server")
        self.server_commands_info = c.boolget("server-commands-info")
        self.server_commands_signals = c.strtupleget("server-commands-signals")
        self.server_readonly = c.boolget("readonly")
        if self.server_readonly and not self.readonly:
            log.info("server is read only")
            self.readonly = True
        if not self.server_keyboard and self.keyboard_helper:
            # swallow packets:
            self.keyboard_helper.send = noop

        i = platform_name(self._remote_platform,
                          c.strtupleget("platform.linux_distribution") or c.strget("platform.release"))
        r = ".".join(str(x) for x in self._remote_version)
        if self._remote_revision:
            r += f"-r{self._remote_revision}"
        mode = c.strget("server.mode", "server")
        bits = c.intget("python.bits", 0)
        bitsstr = "" if bits == 0 else f" {bits}-bit"
        log.info(f"Xpra {mode} server version {std(r)}{bitsstr}")
        if i and i != "unknown":
            log.info(f" running on {std(i)}")
        if c.boolget("desktop") or c.boolget("shadow"):
            v = c.intpair("actual_desktop_size")
            if v:
                w, h = v
                ss = c.tupleget("screen_sizes")
                log.info(f" remote desktop size is {w}x{h}")
                if ss:
                    log_screen_sizes(w, h, ss)
        proxy = c.get("proxy")
        if proxy:
            if isinstance(proxy, dict):
                pcaps = typedict(proxy)
                prefix = ""
            else:
                pcaps = c
                prefix = "proxy."
            proxy_hostname = pcaps.strget(f"{prefix}hostname")
            proxy_platform = pcaps.strget(f"{prefix}platform")
            proxy_release = pcaps.strget(f"{prefix}platform.release")
            proxy_version = pcaps.strget(f"{prefix}version")
            proxy_version = pcaps.strget(f"{prefix}build.version", proxy_version)
            proxy_distro = pcaps.strget(f"{prefix}linux_distribution")
            msg = "via: %s proxy version %s" % (
                platform_name(proxy_platform, proxy_distro or proxy_release),
                std(proxy_version or "unknown")
            )
            if proxy_hostname:
                msg += " on '%s'" % std(proxy_hostname)
            log.info(msg)
        return True

    def process_ui_capabilities(self, caps: typedict) -> None:
        for c in CLIENT_BASES:
            if c != XpraClientBase:
                c.process_ui_capabilities(self, caps)
        # keyboard:
        if self.keyboard_helper:
            modifier_keycodes = caps.dictget("modifier_keycodes", {})
            if modifier_keycodes:
                self.keyboard_helper.set_modifier_mappings(modifier_keycodes)
        self.key_repeat_delay, self.key_repeat_interval = caps.intpair("key_repeat", (-1, -1))
        self.handshake_complete()

    def _process_startup_complete(self, packet: PacketType) -> None:
        log("all the existing windows and system trays have been received")
        super()._process_startup_complete(packet)
        gui_ready()
        if self.tray:
            self.tray.ready()
        self.send_info_request()
        msg = "running"
        try:
            windows = tuple(self._id_to_window.values())
        except AttributeError:
            pass
        else:
            trays = sum(1 for w in windows if w.is_tray())
            wins = sum(1 for w in windows if not w.is_tray())
            if wins:
                msg += f", {wins} windows"
            if trays:
                msg += f", {trays} tray"
        log.info(msg)

    def _process_new_window(self, packet: PacketType):
        window = super()._process_new_window(packet)
        screen_mode = any(self._remote_server_mode.find(x) >= 0 for x in ("desktop", "monitor", "shadow"))
        if self.desktop_fullscreen and screen_mode:
            Gdk = gi_import("Gdk")
            screen = Gdk.Screen.get_default()
            n = screen.get_n_monitors()
            monitor = (len(self._id_to_window) - 1) % n
            window.fullscreen_on_monitor(screen, monitor)
            log("fullscreen_on_monitor: %i", monitor)
        return window

    def handshake_complete(self) -> None:
        oh = self._on_handshake
        self._on_handshake = None
        for cb, args in oh:
            with log.trap_error("Error processing handshake callback %s", cb):
                cb(*args)

    def after_handshake(self, cb: Callable, *args) -> None:
        log("after_handshake(%s, %s) on_handshake=%s", cb, args, Ellipsizer(self._on_handshake))
        if self._on_handshake is None:
            # handshake has already occurred, just call it:
            GLib.idle_add(cb, *args)
        else:
            self._on_handshake.append((cb, args))

    ######################################################################
    # server messages:
    # noinspection PyMethodMayBeStatic
    def _process_server_event(self, packet: PacketType) -> None:
        log(": ".join(str(x) for x in packet[1:]))

    def on_server_setting_changed(self, setting: str, cb: Callable[[str, Any], None]) -> None:
        self._on_server_setting_changed.setdefault(setting, []).append(cb)
        # has the value already been updated:
        value = self.server_setting_updates.get(setting)
        log("on_server_setting_changed%s value=%s", (setting, cb), value)
        if value is not None:
            cb(setting, value)

    def _process_setting_change(self, packet: PacketType) -> None:
        setting = str(packet[1])
        value = packet[2]
        # convert "hello" / "setting" variable names to client variables:
        if setting in (
            "clipboard-limits",
        ):
            # FIXME: this should update the limits?
            pass
        elif setting in (
                "bell", "randr", "cursors", "notifications", "clipboard",
                "clipboard-direction", "session_name",
                "sharing", "sharing-toggle", "lock", "lock-toggle",
                "start-new-commands", "client-shutdown", "webcam",
                "bandwidth-limit", "clipboard-limits",
                "xdg-menu", "monitors",
                "ibus-layouts",
        ):
            setattr(self, "server_%s" % setting.replace("-", "_"), value)
        else:
            log.info("unknown server setting changed: %s=%s", setting, repr_ellipsized(value))
            return
        log("_process_setting_change: %s=%s", setting, Ellipsizer(value))
        # these are too big to log
        if setting not in ("xdg-menu", "monitors", "ibus-layouts"):
            log.info("server setting changed: %s=%s", setting, repr_ellipsized(value))
        self.server_setting_changed(setting, value)

    def server_setting_changed(self, setting: str, value) -> None:
        self.server_setting_updates[setting] = value
        cbs: Sequence[Callable[[str, Any], None]] = self._on_server_setting_changed.get(setting, ())
        log("setting_changed(%s, %s) callbacks=%s", setting, Ellipsizer(value, limit=200), cbs)
        for cb in cbs:
            log("setting_changed(%s, %s) calling %s", setting, Ellipsizer(value, limit=200), cb)
            cb(setting, value)

    def add_control_commands(self) -> None:
        super().add_control_commands()
        try:
            from xpra.net.control.common import ControlCommand, ArgsControlCommand
        except ImportError as e:
            log(f"control commands are not available: {e}")
        else:
            self.control_commands |= {
                "show_session_info": ControlCommand("show-session-info", "Shows the session info dialog", self.show_session_info),
                "show_bug_report": ControlCommand("show-bug-report", "Shows the bug report dialog", self.show_bug_report),
                "name": ArgsControlCommand("name", "Sets the server session name", self.set_server_session_name,
                                           min_args=1, max_args=1),
            }

    def set_server_session_name(self, name: str):
        self.server_session_name = name
        log.info("session name updated from server: %s", self.server_session_name)

    def may_notify_audio(self, summary: str, body: str) -> None:
        self.may_notify(NotificationID.AUDIO, summary, body, icon_name="audio")

    ######################################################################
    # features:
    def send_sharing_enabled(self) -> None:
        assert self.server_sharing and self.server_sharing_toggle
        self.send("sharing-toggle", self.client_supports_sharing)

    def send_lock_enabled(self) -> None:
        assert self.server_lock_toggle
        self.send("lock-toggle", self.client_lock)

    def send_notify_enabled(self) -> None:
        assert self.client_supports_notifications, "cannot toggle notifications: the feature is disabled by the client"
        self.send("set-notify", self.notifications_enabled)

    def send_bell_enabled(self) -> None:
        assert self.client_supports_bell, "cannot toggle bell: the feature is disabled by the client"
        assert self.server_bell, "cannot toggle bell: the feature is disabled by the server"
        self.send("set-bell", self.bell_enabled)

    def send_cursors_enabled(self) -> None:
        assert self.client_supports_cursors, "cannot toggle cursors: the feature is disabled by the client"
        assert self.server_cursors, "cannot toggle cursors: the feature is disabled by the server"
        self.send("set-cursors", self.cursors_enabled)

    def send_force_ungrab(self, wid: int) -> None:
        self.send("force-ungrab", wid)

    def send_keyboard_sync_enabled_status(self, *_args) -> None:
        self.send("set-keyboard-sync-enabled", self.keyboard_sync)

    ######################################################################
    # keyboard:
    def get_keyboard_caps(self) -> dict[str, Any]:
        caps = {}
        kh = self.keyboard_helper
        if self.readonly or not kh:
            # don't bother sending keyboard info, as it won't be used
            caps["keyboard"] = False
        else:
            caps["keyboard"] = True
            caps["ibus"] = True
            caps["modifiers"] = self.get_current_modifiers()
            skip = ("keycodes", "x11_keycodes") if DELAY_KEYBOARD_DATA else ()
            caps["keymap"] = kh.get_keymap_properties(skip)
            # show the user a summary of what we have detected:
            self.keyboard_helper.log_keyboard_info()
            delay_ms, interval_ms = kh.key_repeat_delay, kh.key_repeat_interval
            if delay_ms > 0 and interval_ms > 0:
                caps["key_repeat"] = (delay_ms, interval_ms)
            else:
                # cannot do keyboard_sync without a key repeat value!
                # (maybe we could just choose one?)
                kh.keyboard_sync = False
            caps["keyboard_sync"] = kh.sync
        log("keyboard capabilities: %s", caps)
        return caps

    def next_keyboard_layout(self, update_platform_layout) -> None:
        if self.keyboard_helper:
            self.keyboard_helper.next_layout(update_platform_layout)

    def window_keyboard_layout_changed(self, window=None) -> None:
        # win32 can change the keyboard mapping per window...
        keylog("window_keyboard_layout_changed(%s)", window)
        if self.keyboard_helper:
            self.keyboard_helper.keymap_changed()

    def handle_key_action(self, window, key_event: KeyEvent) -> bool:
        kh = self.keyboard_helper
        if not kh:
            return False
        wid = self._window_to_id[window]
        keylog(f"handle_key_action({window}, {key_event}) wid={wid}")
        if kh.key_handled_as_shortcut(window, key_event.keyname, key_event.modifiers, key_event.pressed):
            return False
        if self.readonly:
            return False
        kh.process_key_event(wid, key_event)
        return False

    def mask_to_names(self, mask) -> list[str]:
        if self.keyboard_helper is None:
            return []
        return self.keyboard_helper.mask_to_names(int(mask))

    ######################################################################
    # windows overrides
    def cook_metadata(self, _new_window, metadata: dict) -> typedict:
        # convert to a typedict and apply client-side overrides:
        tdmeta = typedict(metadata)
        if self.server_is_desktop and self.desktop_fullscreen:
            # force it fullscreen:
            metadata.pop("size-constraints", None)
            metadata["fullscreen"] = True
            # FIXME: try to figure out the monitors we go fullscreen on for X11:
            # if POSIX:
            #    metadata["fullscreen-monitors"] = [0, 1, 0, 1]
        return tdmeta

    ######################################################################
    # network and status:
    def server_connection_state_change(self) -> None:
        if not self._server_ok and hasattr(self, "redraw_spinners"):
            log.info("server is not responding, drawing spinners over the windows")

            def timer_redraw() -> bool:
                if self._protocol is None:
                    # no longer connected!
                    return False
                ok = self.server_ok()
                self.redraw_spinners()
                if ok:
                    log.info("server is OK again")
                return not ok  # repaint again until ok

            GLib.idle_add(self.redraw_spinners)
            GLib.timeout_add(250, timer_redraw)

    ######################################################################
    # packets:
    def init_authenticated_packet_handlers(self) -> None:
        log("init_authenticated_packet_handlers()")
        for c in CLIENT_BASES:
            c.init_authenticated_packet_handlers(self)
        # run from the UI thread:
        self.add_packets("startup-complete", "setting-change", "control", main_thread=True)
        # run directly from the network thread:
        self.add_packets("server-event")

    def process_packet(self, proto, packet) -> None:
        self.check_server_echo(0)
        super().process_packet(proto, packet)



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import struct

from xpra.net.rfb.const import RFBEncoding
from xpra.codecs.rgb_transform import rgb_reformat
from xpra.codecs.pillow.encoder import encode
from xpra.util.objects import typedict
from xpra.util.str_fn import hexstr
from xpra.log import Logger

log = Logger("rfb")

PILLOW_OPTIONS = typedict({"alpha": False})


def pillow_encode(encoding, img):
    return encode(encoding, img, PILLOW_OPTIONS)[1].data


def make_header(encoding, x, y, w, h):
    fbupdate = struct.pack(b"!BBH", 0, 0, 1)
    rect = struct.pack(b"!HHHHi", x, y, w, h, encoding)
    return fbupdate + rect


def rgb222_encode(window, x, y, w, h):
    img = window.get_image(x, y, w, h)
    window.acknowledge_changes()
    header = make_header(RFBEncoding.RAW, x, y, w, h)
    if img.get_pixel_format() != "BGRX":
        log.warn("Warning: cannot convert %s to rgb222", img.get_pixel_format())
        return []
    pixels = img.get_pixels()
    from xpra.codecs.argb.argb import bgra_to_rgb222  # pylint: disable=no-name-in-module
    data = bgra_to_rgb222(pixels)
    return [header, data]


def raw_encode(window, x, y, w, h):
    img = window.get_image(x, y, w, h)
    window.acknowledge_changes()
    header = make_header(RFBEncoding.RAW, x, y, w, h)
    return [header, raw_pixels(img)]


def raw_pixels(img):
    if not img:
        return []
    w = img.get_width()
    h = img.get_height()
    Bpp = len(img.get_pixel_format())  # ie: BGRX -> 4
    if img.get_rowstride() != w * Bpp:
        img.restride(w * Bpp)
    pixels = img.get_pixels()
    assert len(pixels) >= Bpp * w * h, "expected %i pixels for %ix%i-%i but got %i" % (
        Bpp * w * h, w, h, Bpp, len(pixels))
    return pixels[:Bpp * w * h]


def zlib_encode(window, x, y, w, h):
    img = window.get_image(x, y, w, h)
    window.acknowledge_changes()
    if not img:
        return []
    pixels = raw_pixels(img)
    import zlib  # pylint: disable=import-outside-toplevel
    if isinstance(pixels, memoryview):
        pixels = pixels.tobytes()
    data = zlib.compress(pixels, 1)
    log("zlib compressed %i down to %i", len(pixels), len(data))
    header = make_header(RFBEncoding.ZLIB, x, y, w, h) + struct.pack(b"!I", len(data))
    return [header, data]


def tight_encode(window, x, y, w, h, quality=0):
    img = window.get_image(x, y, w, h)
    window.acknowledge_changes()
    if not img:
        return []
    if quality == 10:
        # Fill Compression
        header = make_header(RFBEncoding.TIGHT, x, y, w, h)
        header += struct.pack(b"!B", 0x80)
        pixel_format = img.get_pixel_format()
        log.warn("fill compression of %s", pixel_format)
        if not rgb_reformat(img, ("RGB",), False):
            log.error("Error: cannot convert %s to RGB", pixel_format)
        return [header, raw_pixels(img)]
    # try jpeg:
    data = pillow_encode("jpeg", img)
    header = tight_header(RFBEncoding.TIGHT, x, y, w, h, 0x90, len(data))
    return [header, data]


def tight_header(encoding, x, y, w, h, control, length):
    header = make_header(encoding, x, y, w, h)
    header += struct.pack(b"!B", control)
    # the length header is in a weird format:
    if length < 128:
        header += struct.pack(b"!B", length)
    elif length < 16383:
        header += struct.pack(b"!BB", 0x80 + (length & 0x7F), length >> 7)
    else:
        assert length < 4194303
        header += struct.pack(b"!BBB", 0x80 + (length & 0x7F), 0x80 + ((length >> 7) & 0x7F), length >> 14)
    log("tight header for %i bytes %s", length, hexstr(header))
    return header


def tight_png(window, x, y, w, h):
    img = window.get_image(x, y, w, h)
    window.acknowledge_changes()
    if not img:
        return []
    data = pillow_encode("png", img)
    header = tight_header(RFBEncoding.TIGHT_PNG, x, y, w, h, 0x80 + 0x20, len(data))
    return [header, data]



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any, Optional, Tuple, List, Dict
from xpra.gtk.error import XError

from xpra.x11.bindings.xlib cimport (
    Display, Drawable, Visual, Window, Bool, Pixmap, XID, Status, Atom, Time, CurrentTime, Cursor, XPointer,
    VisualID, XVisualInfo,
    XGetVisualInfo, VisualIDMask, XDefaultVisual,
    GrabModeAsync, XGrabPointer,
    Expose,
    XRectangle, XEvent, XClassHint,
    XWMHints, XSizeHints,
    XCreateWindow, XDestroyWindow, XIfEvent, PropertyNotify,
    XSetWindowAttributes,
    XWindowAttributes, XWindowChanges,
    XDefaultRootWindow,
    XInternAtom, XFree, XGetErrorText,
    XGetSelectionOwner, XSetSelectionOwner, XConvertSelection,
    XMapWindow, XMapRaised, XUnmapWindow, XWithdrawWindow, XReparentWindow, XIconifyWindow, XRaiseWindow,
    NextRequest, XSendEvent, XSelectInput, XAddToSaveSet, XRemoveFromSaveSet,
    XGetWindowAttributes, XGetWindowProperty, XDeleteProperty, XChangeProperty,
    XGetWMNormalHints, XSetWMNormalHints, XGetWMHints, XGetWMProtocols,
    XGetGeometry, XTranslateCoordinates, XConfigureWindow,
    XMoveResizeWindow, XMoveWindow, XResizeWindow,
    XGetInputFocus, XSetInputFocus,
    XAllocClassHint, XAllocSizeHints, XGetClassHint, XSetClassHint,
    XAllocIconSize, XIconSize, XSetIconSizes,
    XQueryTree,
    XKillClient,
)
from libc.stdint cimport uintptr_t
from libc.stdlib cimport free, malloc        # pylint: disable=syntax-error
from libc.string cimport memset

from xpra.log import Logger
log = Logger("x11", "bindings", "window")

###################################
# Headers, python magic
###################################

######
# Xlib primitives and constants
######

DEF XNone = 0

DEF isUnmapped = 0

DEF screen_number = 0

cdef extern from "X11/Xlib.h":
    int CWX
    int CWY
    int CWWidth
    int CWHeight
    int InputOnly
    int InputOutput
    int RevertToParent
    int ClientMessage
    int ButtonPress
    int Button1
    int Button2
    int Button3
    int SelectionNotify
    int ConfigureNotify

    int CopyFromParent

    int CWOverrideRedirect
    int CWEventMask
    int CWColormap
    int CWBorderWidth
    int CWSibling
    int CWStackMode

    int NoEventMask
    int KeyPressMask
    int KeyReleaseMask
    int ButtonPressMask
    int ButtonReleaseMask
    int EnterWindowMask
    int LeaveWindowMask
    int PointerMotionMask
    int PointerMotionHintMask
    int Button1MotionMask
    int Button2MotionMask
    int Button3MotionMask
    int Button4MotionMask
    int Button5MotionMask
    int ButtonMotionMask
    int KeymapStateMask
    int ExposureMask
    int VisibilityChangeMask
    int StructureNotifyMask
    int ResizeRedirectMask
    int SubstructureNotifyMask
    int SubstructureRedirectMask
    int FocusChangeMask
    int PropertyChangeMask
    int ColormapChangeMask
    int OwnerGrabButtonMask

    int AnyPropertyType
    int Success
    int PropModeReplace
    int USPosition
    int PPosition
    int USSize
    int PSize
    int PMinSize
    int IsUnmapped
    int XNone
    int PMaxSize
    int PBaseSize
    int PResizeInc
    int PAspect
    int PWinGravity
    int InputHint
    int StateHint
    int IconPixmapHint
    int IconWindowHint
    int IconPositionHint
    int IconMaskHint
    int WindowGroupHint
    int XUrgencyHint
    int WithdrawnState
    int IconicState
    int NormalState
    int NotifyNormal
    int NotifyGrab
    int NotifyUngrab
    int NotifyWhileGrabbed
    int NotifyNonlinearVirtual
    int NotifyAncestor
    int NotifyVirtual
    int NotifyInferior
    int NotifyNonlinear
    int NotifyPointer
    int NotifyPointerRoot
    int NotifyDetailNone


constants: Dict[str, int] = {
    "CWX"               : CWX,
    "CWY"               : CWY,
    "CWWidth"           : CWWidth,
    "CWHeight"          : CWHeight,
    "CurrentTime"       : CurrentTime,
    "IsUnmapped"        : IsUnmapped,
    "InputOnly"         : InputOnly,
    "RevertToParent"    : RevertToParent,
    "ClientMessage"     : ClientMessage,
    "ButtonPress"       : ButtonPress,
    "Button1"           : Button1,
    "Button2"           : Button2,
    "Button3"           : Button3,
    "SelectionNotify"   : SelectionNotify,
    "ConfigureNotify"   : ConfigureNotify,
    "CWBorderWidth"     : CWBorderWidth,
    "CWSibling"         : CWSibling,
    "CWStackMode"       : CWStackMode,
    "AnyPropertyType"   : AnyPropertyType,
    "Success"           : Success,
    "PropModeReplace"   : PropModeReplace,
    "USPosition"        : USPosition,
    "PPosition"         : PPosition,
    "USSize"            : USSize,
    "PSize"             : PSize,
    "PMinSize"          : PMinSize,
    "XNone"             : XNone,
    "PMaxSize"          : PMaxSize,
    "PBaseSize"         : PBaseSize,
    "PResizeInc"        : PResizeInc,
    "PAspect"           : PAspect,
    "PWinGravity"       : PWinGravity,
    "InputHint"         : InputHint,
    "StateHint"         : StateHint,
    "IconPixmapHint"    : IconPixmapHint,
    "IconWindowHint"    : IconWindowHint,
    "IconPositionHint"  : IconPositionHint,
    "IconMaskHint"      : IconMaskHint,
    "WindowGroupHint"   : WindowGroupHint,
    "XUrgencyHint"      : XUrgencyHint,
    "WithdrawnState"    : WithdrawnState,
    "IconicState"       : IconicState,
    "NormalState"       : NormalState,
    "NotifyNormal"      : NotifyNormal,
    "NotifyGrab"        : NotifyGrab,
    "NotifyUngrab"      : NotifyUngrab,
    "NotifyWhileGrabbed" : NotifyWhileGrabbed,
    "NotifyNonlinear"   : NotifyNonlinear,
    "NotifyNonlinearVirtual" : NotifyNonlinearVirtual,
    "NotifyAncestor"    : NotifyAncestor,
    "NotifyVirtual"     : NotifyVirtual,
    "NotifyInferior"    : NotifyInferior,
    "NotifyNonlinearVirtual" : NotifyNonlinearVirtual,
    "NotifyPointer"     : NotifyPointer,
    "NotifyPointerRoot" : NotifyPointerRoot,
    "NotifyDetailNone"  : NotifyDetailNone,

    "NoEventMask"       : NoEventMask,
    "StructureNotifyMask" : StructureNotifyMask,
    "SubstructureNotifyMask"   : SubstructureNotifyMask,
    "SubstructureRedirectMask" : SubstructureRedirectMask,
    "FocusChangeMask"   : FocusChangeMask,
    "KeyPressMask"      : KeyPressMask,
    "KeyReleaseMask"    : KeyReleaseMask,
    "ButtonPress"       : ButtonPressMask,
    "ButtonReleaseMask" : ButtonReleaseMask,
    "EnterWindowMask"   : EnterWindowMask,
    "LeaveWindowMask"   : LeaveWindowMask,
    "PointerMotionMask" : PointerMotionMask,
    "PointerMotionHintMask": PointerMotionHintMask,
    "Button1MotionMask" : Button1MotionMask,
    "Button2MotionMask" : Button2MotionMask,
    "Button3MotionMask" : Button3MotionMask,
    "Button4MotionMask" : Button4MotionMask,
    "Button5MotionMask" : Button5MotionMask,
    "ButtonMotionMask"  : ButtonMotionMask,
    "KeymapStateMask"   : KeymapStateMask,
    "ExposureMask"      : ExposureMask,
    "VisibilityChangeMask": VisibilityChangeMask,
    "StructureNotifyMask": StructureNotifyMask,
    "ResizeRedirectMask": ResizeRedirectMask,
    "SubstructureNotifyMask": SubstructureNotifyMask,
    "SubstructureRedirectMask": SubstructureRedirectMask,
    "FocusChangeMask"   : FocusChangeMask,
    "PropertyChangeMask": PropertyChangeMask,
    "ColormapChangeMask": ColormapChangeMask,
    "OwnerGrabButtonMask": OwnerGrabButtonMask,
}

MASKS: Dict[int, str] = {}
for name, constant in constants.items():
    if name.endswith("Mask") and constant>0:
        MASKS[constant] = name[:-4]


def get_mask_strs(int mask) -> list[str]:
    masks = []
    for constant, name in MASKS.items():
        if mask & constant:
            masks.append(name)
    return masks


###################################
# Composite
###################################

cdef extern from "X11/extensions/Xcomposite.h":
    Bool XCompositeQueryExtension(Display *, int *, int *)
    Status XCompositeQueryVersion(Display *, int * major, int * minor)
    unsigned int CompositeRedirectManual
    unsigned int CompositeRedirectAutomatic
    void XCompositeRedirectWindow(Display *, Window, int mode)
    void XCompositeRedirectSubwindows(Display *, Window, int mode)
    void XCompositeUnredirectWindow(Display *, Window, int mode)
    void XCompositeUnredirectSubwindows(Display *, Window, int mode)
    Window XCompositeGetOverlayWindow(Display *dpy, Window window)
    void XCompositeReleaseOverlayWindow(Display *dpy, Window window)


cdef extern from "X11/extensions/shape.h":
    Bool XShapeQueryExtension(Display *display, int *event_base, int *error_base)
    Status XShapeQueryVersion(Display *display, int *major_version, int *minor_version)
    Status XShapeQueryExtents(Display *display, Window window, Bool *bounding_shaped, int *x_bounding, int *y_bounding, unsigned int *w_bounding, unsigned int *h_bounding, Bool *clip_shaped, int *x_clip, int *y_clip, unsigned int *w_clip, unsigned int *h_clip)
    void XShapeSelectInput(Display *display, Window window, unsigned long mask)
    unsigned long XShapeInputSelected(Display *display, Window window)
    XRectangle *XShapeGetRectangles(Display *display, Window window, int kind, int *count, int *ordering)

    void XShapeCombineRectangles(Display *display, Window dest, int dest_kind, int x_off, int y_off, XRectangle *rectangles, int n_rects, int op, int ordering)

    cdef int ShapeBounding
    cdef int ShapeClip
    cdef int ShapeInput

SHAPE_KIND: Dict[int, str] = {
    ShapeBounding   : "Bounding",
    ShapeClip       : "Clip",
    ShapeInput      : "ShapeInput",
}

###################################
# Xfixes: cursor events
###################################
cdef extern from "X11/extensions/xfixeswire.h":
    unsigned long XFixesDisplayCursorNotifyMask
    void XFixesSelectCursorInput(Display *, Window w, long mask)

    unsigned int XFixesSetSelectionOwnerNotifyMask
    unsigned int XFixesSelectionWindowDestroyNotifyMask
    unsigned int XFixesSelectionClientCloseNotifyMask

cdef extern from "X11/extensions/Xfixes.h":
    ctypedef struct XFixesCursorNotify:
        char* subtype
        Window XID
        int cursor_serial
        int time
        char* cursor_name
    ctypedef struct XFixesCursorImage:
        short x
        short y
        unsigned short width
        unsigned short height
        unsigned short xhot
        unsigned short yhot
        unsigned long cursor_serial
        unsigned long* pixels
        Atom atom
        char* name
    ctypedef struct XFixesCursorNotifyEvent:
        int type
        unsigned long serial
        Bool send_event
        Display *display
        Window window
        int subtype
        unsigned long cursor_serial
        Time timestamp
        Atom cursor_name

    Bool XFixesQueryExtension(Display *, int *event_base, int *error_base)
    XFixesCursorImage* XFixesGetCursorImage(Display *)

    ctypedef XID XserverRegion

    XserverRegion XFixesCreateRegion(Display *dpy, XRectangle *rectangles, int nrectangles)
    void XFixesDestroyRegion(Display *dpy, XserverRegion region)

    void XFixesSetWindowShapeRegion(Display *dpy, Window win, int shape_kind, int x_off, int y_off, XserverRegion region)

    void XFixesSelectSelectionInput(Display *dpy, Window win, Atom selection, unsigned long eventMask)


###################################
# Xdamage
###################################

cdef extern from "X11/extensions/Xdamage.h":
    ctypedef XID Damage
    unsigned int XDamageReportDeltaRectangles
    #unsigned int XDamageReportRawRectangles
    unsigned int XDamageNotify
    ctypedef struct XDamageNotifyEvent:
        Damage damage
        int level
        Bool more
        XRectangle area
    Bool XDamageQueryExtension(Display *, int * event_base, int *)
    Status XDamageQueryVersion(Display *, int * major, int * minor)
    Damage XDamageCreate(Display *, Drawable, int level)
    void XDamageDestroy(Display *, Damage)
    void XDamageSubtract(Display *, Damage, XserverRegion repair, XserverRegion parts)


cdef inline long cast_to_long(i):
    if i < 0:
        return <long>i
    else:
        return <long><unsigned long>i


class PropertyError(Exception):
    pass
class BadPropertyType(PropertyError):
    pass


class PropertyOverflow(PropertyError):
    pass


from xpra.x11.bindings.core cimport X11CoreBindingsInstance

cdef int CONFIGURE_GEOMETRY_MASK = CWX | CWY | CWWidth | CWHeight


cdef class X11WindowBindingsInstance(X11CoreBindingsInstance):

    cdef object has_xshape

    def __cinit__(self):
        self.has_xshape = None

    def __repr__(self):
        return "X11WindowBindings(%s)" % self.display_name

    ###################################
    # Extension testing
    ###################################

    # X extensions all have different APIs for negotiating their
    # availability/version number, but a number of the more recent ones are
    # similar enough to share code (in particular, Composite and DAMAGE, and
    # probably also Xfixes, Xrandr, etc.).  (But note that we don't actually have
    # to query for Xfixes support because 1) any server that can handle us at all
    # already has a sufficiently advanced version of Xfixes, and 2) GTK+ already
    # enables Xfixes for us automatically.)
    cdef ensure_extension_support(self, major, minor, extension,
                                   Bool (*query_extension)(Display*, int*, int*),
                                   Status (*query_version)(Display*, int*, int*)):
        cdef int event_base = 0, ignored = 0
        if not (query_extension)(self.display, &event_base, &ignored):
            raise ValueError("X11 extension %s not available" % extension)
        log("X11 extension %s event_base=%i", extension, event_base)
        cdef int cmajor = major, cminor = minor
        if (query_version)(self.display, &cmajor, &cminor):
            # See X.org bug #14511:
            log("found X11 extension %s with version %i.%i", extension, major, minor)
            if cmajor<major or (cmajor==major and cminor<minor):
                raise ValueError("%s v%i.%i not supported; required: v%i.%i"
                                 % (extension, cmajor, cminor, major, minor))


    def get_all_x11_windows(self) -> List[Window]:
        cdef Window root = XDefaultRootWindow(self.display);
        return self.get_all_children(root)

    def get_all_children(self, Window xid) -> List[Window]:
        self.context_check("get_all_children")
        cdef Window root = XDefaultRootWindow(self.display)
        cdef Window parent = 0
        cdef Window * children = <Window *> 0
        cdef unsigned int i, nchildren = 0
        windows = []
        try:
            if not XQueryTree(self.display,
                              xid,
                              &root, &parent, &children, &nchildren):
                return []
            for i in range(nchildren):
                windows.append(children[i])
        finally:
            if nchildren > 0 and children != NULL:
                XFree(children)
        for window in tuple(windows):
            windows += self.get_all_children(window)
        return windows

    def get_absolute_position(self, Window xid) -> Tuple[int, int]:
        self.context_check("get_absolute_position")
        cdef Window root = XDefaultRootWindow(self.display)
        cdef int dest_x = 0, dest_y = 0
        cdef Window child = 0
        if not XTranslateCoordinates(self.display, xid,
                                     root,
                                     0, 0,
                                     &dest_x, &dest_y, &child):
            return None
        return dest_x, dest_y

    def MapWindow(self, Window xwindow) -> None:
        self.context_check("MapWindow")
        XMapWindow(self.display, xwindow)

    def MapRaised(self, Window xwindow) -> None:
        self.context_check("MapRaised")
        XMapRaised(self.display, xwindow)

    def Withdraw(self, Window xwindow) -> None:
        self.context_check("Withdraw")
        return XWithdrawWindow(self.display, xwindow, screen_number)

    def Reparent(self, Window xwindow, Window xparent, int x, int y) -> None:
        self.context_check("Reparent")
        XReparentWindow(self.display, xwindow, xparent, x, y)

    def Iconify(self, Window xwindow) -> None:
        self.context_check("Iconify")
        return XIconifyWindow(self.display, xwindow, screen_number)

    ###################################
    # XUnmapWindow
    ###################################
    def Unmap(self, Window xwindow) -> long:
        self.context_check("Unmap")
        cdef unsigned long serial = NextRequest(self.display)
        XUnmapWindow(self.display, xwindow)
        return serial

    def getWindowAttributes(self, Window xwindow) -> Dict[str, Any]:
        self.context_check("getWindowAttributes")
        cdef XWindowAttributes attrs
        cdef Status status = XGetWindowAttributes(self.display, xwindow, &attrs)
        if status==0:
            return None
        return {
            "geometry"  : (attrs.x, attrs.y, attrs.width, attrs.height, attrs.border_width),
            "depth"     : attrs.depth,
            "visual"    : {
                "visual-id"     : attrs.visual.visualid,
                #"class"         : attrs.visual.c_class,
                "red-mask"      : attrs.visual.red_mask,
                "green-mask"    : attrs.visual.green_mask,
                "blue-mask"     : attrs.visual.blue_mask,
                "bits-per-rgb"  : attrs.visual.bits_per_rgb,
                "map_entries"   : attrs.visual.map_entries,
            },
            "bit-gravity" : attrs.bit_gravity,
            "win-gravity" : attrs.win_gravity,
            "backing-store" : attrs.backing_store,
            "backing-planes" : attrs.backing_planes,
            "backing-pixel" : attrs.backing_pixel,
            "save-under"    : bool(attrs.save_under),
            #"colormap"  : 0,
            "map-installed" : attrs.map_installed,
            "map-state" : attrs.map_state,
            "all-events-mask" : attrs.all_event_masks,
            "your_event_mask"   : attrs.your_event_mask,
            "do-not-propagate-mask" : attrs.do_not_propagate_mask,
            "override-redirect"     : attrs.override_redirect,
        }

    def getEventMask(self, Window xwindow) -> long:
        self.context_check("getEventMask")
        cdef XWindowAttributes attrs
        cdef Status status = XGetWindowAttributes(self.display, xwindow, &attrs)
        if status==0:
            return 0
        return attrs.your_event_mask;

    def setEventMask(self, Window xwindow, int mask) -> None:
        self.context_check("setEventMask")
        XSelectInput(self.display, xwindow, mask)

    # Mapped status
    def is_mapped(self, Window xwindow) -> bool:
        return self.get_map_state(xwindow) != IsUnmapped

    def get_map_state(self, Window xwindow) -> int:
        self.context_check("get_map_state")
        cdef XWindowAttributes attrs
        cdef Status status = XGetWindowAttributes(self.display, xwindow, &attrs)
        if status==0:
            return isUnmapped
        return attrs.map_state

    # Override-redirect status
    def is_override_redirect(self, Window xwindow) -> bool:
        self.context_check("is_override_redirect")
        cdef XWindowAttributes or_attrs
        cdef Status status = XGetWindowAttributes(self.display, xwindow, &or_attrs)
        if status==0:
            return False
        return bool(or_attrs.override_redirect)

    # Mapped status
    def is_inputonly(self, Window xwindow) -> bool:
        self.context_check("is_inputonly")
        cdef XWindowAttributes attrs
        cdef Status status = XGetWindowAttributes(self.display, xwindow, &attrs)
        if status==0:
            return False
        return bool(attrs._class==InputOnly)


    def geometry_with_border(self, Window xwindow) -> Tuple[int, int, int, int, int] | None:
        self.context_check("geometry_with_border")
        cdef XWindowAttributes geom_attrs
        cdef Status status = XGetWindowAttributes(self.display, xwindow, &geom_attrs)
        if status==0:
            return None
        return (geom_attrs.x, geom_attrs.y, geom_attrs.width, geom_attrs.height, geom_attrs.border_width)

    def get_depth(self, Drawable d) -> int:
        self.context_check("get_depth")
        cdef Window root
        cdef int x, y
        cdef unsigned int width, height, border_width, depth
        if not XGetGeometry(self.display, d, &root,
                            &x, &y, &width, &height, &border_width, &depth):
            return 0
        return depth

    # Focus management
    def XSetInputFocus(self, Window xwindow, Time time=CurrentTime) -> None:
        self.context_check("XSetInputFocus")
        # Always does RevertToParent
        XSetInputFocus(self.display, xwindow, RevertToParent, time)

    def XGetInputFocus(self) -> Tuple[int, int]:
        #context check not needed!
        cdef Window w = 0
        cdef int revert_to = 0
        XGetInputFocus(self.display, &w, &revert_to)
        return int(w), int(revert_to)

    def XRaiseWindow(self, Window window) -> None:
        self.context_check("XRaiseWindow")
        XRaiseWindow(self.display, window)

    ###################################
    # XKillClient
    ###################################
    def XKillClient(self, Window xwindow) -> int:
        self.context_check("XKillClient")
        return XKillClient(self.display, xwindow)

    ###################################
    # Shape
    ###################################
    def displayHasXShape(self) -> bool:
        cdef int event_base = 0, ignored = 0
        cdef int cmajor, cminor
        if self.has_xshape is not None:
            pass
        elif not XShapeQueryExtension(self.display, &event_base, &ignored):
            log.warn("X11 extension XShape not available")
            self.has_xshape = False
        else:
            log("X11 extension XShape event_base=%i", event_base)
            if not XShapeQueryVersion(self.display, &cmajor, &cminor):
                log.warn("XShape version query failed")
                self.has_xshape = False
            else:
                log("found X11 extension XShape with version %i.%i", cmajor, cminor)
                self.has_xshape = True
        log("displayHasXShape()=%s", self.has_xshape)
        return self.has_xshape

    def XShapeSelectInput(self, Window window) -> None:
        self.context_check("XShapeSelectInput")
        cdef int ShapeNotifyMask = 1
        XShapeSelectInput(self.display, window, ShapeNotifyMask)

    def XShapeQueryExtents(self, Window window) -> Tuple[Tuple, Tuple]:
        self.context_check("XShapeQueryExtents")
        cdef Bool bounding_shaped, clip_shaped
        cdef int x_bounding, y_bounding, x_clip, y_clip
        cdef unsigned int w_bounding, h_bounding, w_clip, h_clip
        if not XShapeQueryExtents(self.display, window,
                                  &bounding_shaped, &x_bounding, &y_bounding, &w_bounding, &h_bounding,
                                  &clip_shaped, &x_clip, &y_clip, &w_clip, &h_clip):
            return None
        return (
            (bounding_shaped, x_bounding, y_bounding, w_bounding, h_bounding),
            (clip_shaped, x_clip, y_clip, w_clip, h_clip)
        )

    def XShapeGetRectangles(self, Window window, int kind) -> List[Tuple[int, int, int, int]]:
        self.context_check("XShapeGetRectangles")
        cdef int count, ordering
        cdef XRectangle* rect = XShapeGetRectangles(self.display, window, kind, &count, &ordering)
        if rect==NULL or count<=0:
            return []
        rectangles = []
        cdef int i
        for i in range(count):
            rectangles.append((rect[i].x, rect[i].y, rect[i].width, rect[i].height))
        return rectangles

    def XShapeCombineRectangles(self, Window window, int kind, int x_off, int y_off, rectangles) -> None:
        self.context_check("XShapeCombineRectangles")
        cdef int n_rects = len(rectangles)
        cdef int op = 0     #SET
        cdef int ordering = 0   #Unsorted
        cdef size_t l = sizeof(XRectangle) * n_rects
        cdef XRectangle *rects = <XRectangle*> malloc(l)
        if rects==NULL:
            raise RuntimeError("failed to allocate %i bytes of memory for xshape rectangles" % l)
        cdef int i = 0
        for r in rectangles:
            rects[i].x = r[0]
            rects[i].y = r[1]
            rects[i].width = r[2]
            rects[i].height = r[3]
            i += 1
        XShapeCombineRectangles(self.display, window, kind, x_off, y_off,
                                rects, n_rects, op, ordering)
        free(rects)


    ###################################
    # Composite
    ###################################
    def ensure_XComposite_support(self) -> None:
        # We need NameWindowPixmap, but we don't need the overlay window
        # (v0.3) or the special manual-redirect clipping semantics (v0.4).
        self.ensure_extension_support(0, 2, "Composite",
                                  XCompositeQueryExtension,
                                  XCompositeQueryVersion)

    def displayHasXComposite(self) -> bool:
        try:
            self.ensure_XComposite_support()
            return  True
        except Exception as e:
            log.error("%s", e)
        return False

    def XCompositeRedirectWindow(self, Window xwindow) -> None:
        self.context_check("XCompositeRedirectWindow")
        XCompositeRedirectWindow(self.display, xwindow, CompositeRedirectManual)

    def XCompositeRedirectSubwindows(self, Window xwindow) -> None:
        self.context_check("XCompositeRedirectSubwindows")
        XCompositeRedirectSubwindows(self.display, xwindow, CompositeRedirectManual)

    def XCompositeUnredirectWindow(self, Window xwindow) -> None:
        self.context_check("XCompositeUnredirectWindow")
        XCompositeUnredirectWindow(self.display, xwindow, CompositeRedirectManual)

    def XCompositeUnredirectSubwindows(self, Window xwindow) -> None:
        self.context_check("XCompositeUnredirectSubwindows")
        XCompositeUnredirectSubwindows(self.display, xwindow, CompositeRedirectManual)

    def XCompositeGetOverlayWindow(self, Window window) -> Window:
        self.context_check("XCompositeGetOverlayWindow")
        return XCompositeGetOverlayWindow(self.display, window)

    def XCompositeReleaseOverlayWindow(self, Window window) -> None:
        self.context_check("XCompositeReleaseOverlayWindow")
        XCompositeReleaseOverlayWindow(self.display, window)

    def AllowInputPassthrough(self, Window window) -> None:
        self.context_check("AllowInputPassthrough")
        cdef XserverRegion region = XFixesCreateRegion(self.display, NULL, 0)
        XFixesSetWindowShapeRegion(self.display, window, ShapeBounding, 0, 0, 0)
        XFixesSetWindowShapeRegion(self.display, window, ShapeInput, 0, 0, region)
        XFixesDestroyRegion(self.display, region)

    ###################################
    # Xdamage
    ###################################
    def ensure_XDamage_support(self) -> None:
        self.ensure_extension_support(1, 0, "DAMAGE",
                                  XDamageQueryExtension,
                                  XDamageQueryVersion)

    def XDamageCreate(self, Window xwindow) -> None:
        self.context_check("XDamageCreate")
        return XDamageCreate(self.display, xwindow, XDamageReportDeltaRectangles)

    def XDamageDestroy(self, Damage handle) -> None:
        self.context_check("XDamageDestroy")
        XDamageDestroy(self.display, handle)

    def XDamageSubtract(self, Damage handle) -> None:
        self.context_check("XDamageSubtract")
        # def xdamage_acknowledge(display_source, handle, x, y, width, height):
        # cdef XRectangle rect
        # rect.x = x
        # rect.y = y
        # rect.width = width
        # rect.height = height
        # repair = XFixesCreateRegion(display, &rect, 1)
        # XDamageSubtract(display, handle, repair, XNone)
        # XFixesDestroyRegion(display, repair)

        # DeltaRectangles mode + XDamageSubtract is broken, because repair
        # operations trigger a flood of re-reported events (see freedesktop.org bug
        # #14648 for details).  So instead we always repair all damage.  This
        # means we may get redundant damage notifications if areas outside of the
        # rectangle we actually repaired get re-damaged, but it avoids the
        # quadratic blow-up that fixing just the correct area causes, and still
        # reduces the number of events we receive as compared to just using
        # RawRectangles mode.  This is very important for things like, say,
        # drawing a scatterplot in R, which may make hundreds of thousands of
        # draws to the same location, and with RawRectangles mode xpra can lag by
        # seconds just trying to keep track of the damage.
        XDamageSubtract(self.display, handle, XNone, XNone)

    ###################################
    # Smarter convenience wrappers
    ###################################

    def XGetSelectionOwner(self, atom) -> Window:
        self.context_check("XGetSelectionOwner")
        return XGetSelectionOwner(self.display, self.str_to_atom(atom))

    def XSetSelectionOwner(self, Window xwindow, atom, Time time=CurrentTime):
        self.context_check("XSetSelectionOwner")
        return XSetSelectionOwner(self.display, self.str_to_atom(atom), xwindow, time)

    def sendClientMessage(self, Window xtarget, Window xwindow, int propagate, int event_mask,
                          message_type, data0=0, data1=0, data2=0, data3=0, data4=0) -> None:
        self.context_check("sendClientMessage")
        # data0 etc. are passed through get_xatom, so they can be integers, which
        # are passed through directly, or else they can be strings, which are
        # converted appropriately.
        cdef XEvent e
        log("sendClientMessage(%#x, %#x, %#x, %#x, %s, %s, %s, %s, %s, %s)", xtarget, xwindow, propagate, event_mask,
                                        message_type, data0, data1, data2, data3, data4)
        e.type = ClientMessage
        e.xany.display = self.display
        e.xany.window = xwindow
        e.xclient.message_type = self.str_to_atom(message_type)
        e.xclient.format = 32
        e.xclient.data.l[0] = cast_to_long(self.xatom(data0))
        e.xclient.data.l[1] = cast_to_long(self.xatom(data1))
        e.xclient.data.l[2] = cast_to_long(self.xatom(data2))
        e.xclient.data.l[3] = cast_to_long(self.xatom(data3))
        e.xclient.data.l[4] = cast_to_long(self.xatom(data4))
        cdef Status s = XSendEvent(self.display, xtarget, propagate, event_mask, &e)
        if s == 0:
            raise ValueError("failed to serialize ClientMessage")

    def sendClick(self, Window xtarget, int button, onoff, int x_root, int y_root, int x, int y) -> None:
        self.context_check("sendClick")
        cdef Window r = XDefaultRootWindow(self.display)
        log("sending message to %#x", xtarget)
        cdef XEvent e
        e.type = ButtonPress
        e.xany.display = self.display
        e.xany.window = xtarget
        #e.xclient.message_type = self.str_to_atom(message_type)
        e.xclient.format = 32
        if button==1:
            e.xbutton.button = Button1
        elif button==2:
            e.xbutton.button = Button2
        else:
            e.xbutton.button = Button3
        e.xbutton.same_screen = True
        e.xbutton.root = r
        e.xbutton.x_root = x_root
        e.xbutton.y_root = y_root
        e.xbutton.x = x
        e.xbutton.y = y
        e.xbutton.state = int(onoff)
        cdef Status s = XSendEvent(self.display, xtarget, False, 0, &e)
        if s == 0:
            raise ValueError("failed to serialize ButtonPress Message")

    def send_xembed_message(self, Window xwindow, opcode, detail, data1, data2) -> None:
        """
         Display* dpy, /* display */
         Window w, /* receiver */
         long message, /* message opcode */
         long detail  /* message detail */
         long data1  /* message data 1 */
         long data2  /* message data 2 */
         """
        self.context_check("send_xembed_message")
        cdef XEvent e
        e.xany.display = self.display
        e.xany.window = xwindow
        e.xany.type = ClientMessage
        e.xclient.message_type = self.str_to_atom("_XEMBED")
        e.xclient.format = 32
        e.xclient.data.l[0] = CurrentTime
        e.xclient.data.l[1] = opcode
        e.xclient.data.l[2] = detail
        e.xclient.data.l[3] = data1
        e.xclient.data.l[4] = data2
        s = XSendEvent(self.display, xwindow, False, NoEventMask, &e)
        if s == 0:
            raise ValueError("failed to serialize XEmbed Message")

    def send_expose(self, Window xwindow, int x, int y, int width, int height, int count=0):
        cdef XEvent e
        e.xany.display = self.display
        e.xany.window = xwindow
        e.xany.type = Expose
        e.xexpose.x = x
        e.xexpose.y = y
        e.xexpose.width = width
        e.xexpose.height = height
        e.xexpose.count = count
        s = XSendEvent(self.display, xwindow, False, NoEventMask, &e)
        if s == 0:
            raise ValueError("failed to serialize XExpose Message")

    ###################################
    # Clipboard
    ###################################
    def selectXFSelectionInput(self, Window window, selection_str) -> None:
        self.context_check("selectXFSelectionInput")
        cdef unsigned long event_mask = (
            XFixesSetSelectionOwnerNotifyMask |
            XFixesSelectionWindowDestroyNotifyMask |
            XFixesSelectionClientCloseNotifyMask
            )
        cdef Atom selection = self.str_to_atom(selection_str)
        XFixesSelectSelectionInput(self.display, window, selection, event_mask)

    def selectSelectionInput(self, Window xwindow) -> None:
        self.context_check("selectSelectionInput")
        self.addXSelectInput(xwindow, SelectionNotify)

    def sendSelectionNotify(self, Window xwindow, selection, target, property, Time time=CurrentTime) -> None:
        self.context_check("sendSelectionNotify")
        cdef XEvent e
        e.type = SelectionNotify
        e.xselection.requestor = xwindow
        e.xselection.selection = self.str_to_atom(selection)
        e.xselection.target = self.str_to_atom(target)
        e.xselection.time = time
        if property:
            e.xselection.property = self.str_to_atom(property)
        else:
            e.xselection.property = 0
        cdef Status s = XSendEvent(self.display, xwindow, True, 0, &e)
        if s == 0:
            raise ValueError("failed to serialize SelectionNotify")

    def ConvertSelection(self, selection, target, property, Window requestor, Time time=CurrentTime) -> int:
        self.context_check("ConvertSelection")
        return XConvertSelection(self.display, self.str_to_atom(selection), self.str_to_atom(target),
                                 self.str_to_atom(property), requestor, time)

    def sendConfigureNotify(self, Window xwindow) -> None:
        self.context_check("sendConfigureNotify")
        cdef Window root_window = XDefaultRootWindow(self.display)

        # Get basic attributes
        cdef XWindowAttributes attrs
        XGetWindowAttributes(self.display, xwindow, &attrs)

        # Figure out where the window actually is in root coordinate space
        cdef int dest_x = 0, dest_y = 0
        cdef Window child = 0
        if not XTranslateCoordinates(self.display, xwindow,
                                     root_window,
                                     0, 0,
                                     &dest_x, &dest_y, &child):
            # Window seems to have disappeared, so never mind.
            log("couldn't TranslateCoordinates (maybe window is gone)")
            return

        # Send synthetic ConfigureNotify (ICCCM 4.2.3, for example)
        cdef XEvent e
        e.type = ConfigureNotify
        e.xconfigure.event = xwindow
        e.xconfigure.window = xwindow
        e.xconfigure.x = dest_x
        e.xconfigure.y = dest_y
        e.xconfigure.width = attrs.width
        e.xconfigure.height = attrs.height
        e.xconfigure.border_width = attrs.border_width
        e.xconfigure.above = XNone
        e.xconfigure.override_redirect = attrs.override_redirect

        cdef Status s = XSendEvent(self.display, xwindow, False, StructureNotifyMask, &e)
        if s == 0:
            raise ValueError("failed to serialize ConfigureNotify")

    def CreateCorralWindow(self, Window parent, Window xid, int x, int y) -> Window:
        self.context_check("CreateCorralWindow")
        # copy the dimensions from the window we will wrap:
        cdef int ox, oy
        cdef Window root
        cdef unsigned int width, height, border_width, depth
        if not XGetGeometry(self.display, xid, &root,
                        &ox, &oy, &width, &height, &border_width, &depth):
            return 0
        cdef long event_mask = PropertyChangeMask | StructureNotifyMask | SubstructureNotifyMask
        return self.CreateWindow(parent, x, y, width, height, OR=False, event_mask=event_mask)

    def CreateWindow(self, Window parent, int x=0, int y=0, int width=1, int height=1,
                           depth=0, int OR=0, long event_mask=0, unsigned long visualid=0) -> Window:
        self.context_check("CreateWindow")
        cdef XSetWindowAttributes attributes
        memset(<void*> &attributes, 0, sizeof(XSetWindowAttributes))
        attributes.event_mask = event_mask
        cdef unsigned long valuemask = CWEventMask
        if OR:
            valuemask |= CWOverrideRedirect
            attributes.override_redirect = 1
        if depth == 0:
            depth = self.get_depth(parent)
        cdef Visual* visual = <Visual*> CopyFromParent
        if visualid:
            visual = self.get_visual(visualid)
        return XCreateWindow(self.display, parent,
                             x, y, width, height, 0, depth,
                             InputOutput, visual,
                             valuemask, &attributes)

    cdef Visual* get_visual(self, int visualid):
        cdef Visual* visual = NULL
        cdef int count
        cdef XVisualInfo vinfo_template
        vinfo_template.visualid = visualid
        cdef XVisualInfo *vinfo = XGetVisualInfo(self.display, VisualIDMask, &vinfo_template, &count)
        if count != 1 or vinfo == NULL:
            log.error("Error: visual %i not found, count=%i, vinfo=%#x", visualid, count, <uintptr_t> vinfo)
        else:
            visual = vinfo[0].visual
        if vinfo:
            XFree(vinfo)
        return visual

    def get_default_visualid(self) -> int:
        cdef Visual *visual = XDefaultVisual(self.display, 0)
        if not visual:
            return 0
        return visual.visualid

    def get_rgba_visualid(self, depth=32) -> int:
        cdef XVisualInfo vinfo_template
        cdef int count
        cdef XVisualInfo *vinfo = XGetVisualInfo(self.display, 0, &vinfo_template, &count)
        if not count or vinfo == NULL:
            log.error("Error: no visuals found, count=%i, vinfo=%#x", count, <uintptr_t> vinfo)
            if vinfo:
                XFree(vinfo)
            return 0
        cdef unsigned long visualid = 0
        for i in range(count):
            if vinfo[i].depth != depth:
                continue
            # check rgb mask:
            if vinfo[i].red_mask != 0xff0000 or vinfo[i].green_mask != 0x00ff00 or vinfo[i].blue_mask != 0x0000ff:
                continue
            visualid = vinfo[i].visualid
            break
        XFree(vinfo)
        return visualid

    def DestroyWindow(self, Window w) -> None:
        self.context_check("DestroyWindow")
        return XDestroyWindow(self.display, w)

    def ConfigureWindow(self, Window xwindow,
                        int x, int y, int width, int height, int border=0,
                        int sibling=0, int stack_mode=0,
                        int value_mask=CONFIGURE_GEOMETRY_MASK) -> None:
        self.context_check("ConfigureWindow")
        cdef XWindowChanges changes
        changes.x = x
        changes.y = y
        changes.width = width
        changes.height = height
        changes.border_width = border
        changes.sibling = sibling
        changes.stack_mode = stack_mode
        XConfigureWindow(self.display, xwindow, value_mask, &changes)

    def configure(self, Window xwindow, x, y, width, height, fields=None) -> None:
        # Reconfigure the window.  We have to use XConfigureWindow directly
        # instead of GdkWindow.resize, because GDK does not give us any way to
        # squash the border.

        # The caller can pass an XConfigureWindow-style fields mask to turn off
        # some of these bits; this is useful if they are pulling such a field out
        # of a ConfigureRequest (along with the other arguments they are passing
        # to us).  This also means we need to be careful to zero out any bits
        # besides these, because they could be set to anything.
        self.context_check("configure")
        cdef int geom_flags = CWX | CWY | CWWidth | CWHeight
        if fields is None:
            fields = geom_flags
        else:
            fields = fields & geom_flags
        # But we always unconditionally squash the border to zero.
        fields = fields | CWBorderWidth
        self.ConfigureWindow(xwindow, x, y, width, height, value_mask=fields)

    def MoveResizeWindow(self, Window xwindow, int x, int y, int width, int height) -> bool:
        self.context_check("MoveResizeWindow")
        return bool(XMoveResizeWindow(self.display, xwindow, x, y, width, height))

    def ResizeWindow(self, Window xwindow, int width, int height) -> bool:
        self.context_check("ResizeWindow")
        return bool(XResizeWindow(self.display, xwindow, width, height))

    def MoveWindow(self, Window xwindow, int x, int y) -> bool:
        self.context_check("MoveWindow")
        return bool(XMoveWindow(self.display, xwindow, x, y))

    def get_event_mask_strs(self, Window xwindow):
        self.context_check("get_event_mask_strs")
        cdef XWindowAttributes curr
        XGetWindowAttributes(self.display, xwindow, &curr)
        return get_mask_strs(curr.your_event_mask)

    def addDefaultEvents(self, Window xwindow) -> None:
        self.context_check("addDefaultEvents")
        ADDMASK = StructureNotifyMask | PropertyChangeMask | FocusChangeMask | PointerMotionMask | PointerMotionHintMask | ButtonMotionMask
        self.addXSelectInput(xwindow, ADDMASK)

    def addXSelectInput(self, Window xwindow, long add_mask) -> None:
        self.context_check("addXSelectInput")
        cdef XWindowAttributes curr
        XGetWindowAttributes(self.display, xwindow, &curr)
        cdef long mask = curr.your_event_mask | add_mask
        XSelectInput(self.display, xwindow, mask)

    def substructureRedirect(self, Window xwindow) -> None:
        """Enable SubstructureRedirect on the given window.

        This enables reception of MapRequest and ConfigureRequest events.  At the
        X level, it also enables the reception of CirculateRequest events, but
        those are pretty useless, so we just ignore such events unconditionally
        rather than routing them anywhere.  (The circulate request appears to be
        in the protocol just so simple window managers have an easy way to
        implement the equivalent of alt-tab; I can't imagine how it'd be useful
        these days.  Metacity and KWin do not support it; GTK+/GDK and Qt4 provide
        no way to actually send it.)"""
        self.context_check("substructureRedirect")
        self.addXSelectInput(xwindow, SubstructureRedirectMask)

    def selectFocusChange(self, Window xwindow) -> None:
        self.context_check("selectFocusChange")
        self.addXSelectInput(xwindow, FocusChangeMask)

    def XGetWindowProperty(self, Window xwindow, property, req_type="",
                           int buffer_size=64*1024, delete=False, incr=False) -> bytes:
        # NB: Accepts req_type == 0 for AnyPropertyType
        # "64k is enough for anybody"
        # (Except, I've found window icons that are strictly larger)
        self.context_check("XGetWindowProperty")
        cdef Atom xactual_type = <Atom> 0
        cdef int actual_format = 0
        cdef unsigned long nitems = 0, bytes_after = 0
        cdef unsigned char * prop = <unsigned char*> 0
        cdef Atom xreq_type = AnyPropertyType
        if req_type:
            xreq_type = self.str_to_atom(req_type)
        # This is the most bloody awful API I have ever seen.  You will probably
        # not be able to understand this code fully without reading
        # XGetWindowProperty's man page at least 3 times, slowly.
        cdef Status status = XGetWindowProperty(self.display,
                                                xwindow,
                                                self.str_to_atom(property),
                                                0,
                                                # This argument has to be divided by 4.  Thus
                                                # speaks the spec.
                                                buffer_size // 4,
                                                delete,
                                                xreq_type, &xactual_type,
                                                &actual_format, &nitems, &bytes_after, &prop)
        if status != Success:
            raise PropertyError("no such window")
        if xactual_type == XNone:
            return None
        if xreq_type and xreq_type != xactual_type:
            raise BadPropertyType("expected %s but got %s" % (req_type, self.get_atom_name(xactual_type)))
        # This should only occur for bad property types:
        assert not (bytes_after and not nitems)
        if bytes_after and not incr:
            raise PropertyOverflow("reserved %i bytes for %s buffer, but data is bigger by %i bytes!" % (buffer_size, req_type, bytes_after))
        # actual_format is in (8, 16, 32), and is the number of bits in a logical
        # element.  However, this doesn't mean that each element is stored in that
        # many bits, oh no.  On a 32-bit machine it is, but on a 64-bit machine,
        # iff the output array contains 32-bit integers, than each one is given
        # 64-bits of space.
        assert actual_format > 0
        if actual_format == 8:
            bytes_per_item = 1
        elif actual_format == 16:
            bytes_per_item = sizeof(short)
        elif actual_format == 32:
            bytes_per_item = sizeof(long)
        else:
            assert False
        cdef int nbytes = bytes_per_item * nitems
        data = (<char *> prop)[:nbytes]
        XFree(prop)
        return data

    def GetWindowPropertyType(self, Window xwindow, property, incr=False) -> Tuple[str, int]:
        #as above, but for any property type
        #and returns the type found
        self.context_check("GetWindowPropertyType")
        cdef int buffer_size = 64 * 1024
        cdef Atom xactual_type = <Atom> 0
        cdef int actual_format = 0
        cdef unsigned long nitems = 0, bytes_after = 0
        cdef unsigned char * prop = <unsigned char*> 0
        cdef Status status
        cdef Atom xreq_type = AnyPropertyType
        status = XGetWindowProperty(self.display,
                                     xwindow,
                                     self.str_to_atom(property),
                                     0,
                                     # This argument has to be divided by 4.  Thus
                                     # speaks the spec.
                                     buffer_size // 4,
                                     False,
                                     xreq_type, &xactual_type,
                                     &actual_format, &nitems, &bytes_after, &prop)
        if status != Success:
            raise XError("XGetWindowProperty status: %s" % status)
        if xactual_type == XNone:
            raise BadPropertyType("None type")
        # This should only occur for bad property types:
        assert not (bytes_after and not nitems)
        if bytes_after and not incr:
            raise BadPropertyType("incomplete data: %i bytes after" % bytes_after)
        assert actual_format in (8, 16, 32)
        XFree(prop)
        return self.get_atom_name(xactual_type), actual_format

    def XDeleteProperty(self, Window xwindow, property) -> None:
        self.context_check("XDeleteProperty")
        XDeleteProperty(self.display, xwindow, self.str_to_atom(property))

    def XChangeProperty(self, Window xwindow, property, dtype, int dformat, data) -> None:
        "Set a property on a window."
        self.context_check("XChangeProperty")
        if dformat not in (8, 16, 32):
            raise ValueError(f"invalid format for property {property}: {dformat}")
        cdef unsigned char nbytes = dformat//8
        if dformat==32:
            nbytes = sizeof(long)
        assert len(data) % nbytes == 0, "size of data is not a multiple of %s" % nbytes
        cdef int nitems = len(data) // nbytes
        cdef char * data_str = data
        XChangeProperty(self.display, xwindow,
                        self.str_to_atom(property),
                        self.str_to_atom(dtype),
                        dformat,
                        PropModeReplace,
                        <unsigned char *>data_str,
                        nitems)

    # Save set handling
    def XAddToSaveSet(self, Window xwindow) -> None:
        self.context_check("XAddToSaveSet")
        XAddToSaveSet(self.display, xwindow)

    def XRemoveFromSaveSet(self, Window xwindow) -> None:
        self.context_check("XRemoveFromSaveSet")
        XRemoveFromSaveSet(self.display, xwindow)

    def setRootIconSizes(self, int w, int h) -> None:
        cdef Window root = XDefaultRootWindow(self.display);
        cdef XIconSize *icon_size = XAllocIconSize()
        assert icon_size
        icon_size.min_width = w
        icon_size.max_width = w
        icon_size.width_inc = 0
        icon_size.min_height = h
        icon_size.max_height = h
        icon_size.height_inc = 0
        XSetIconSizes(self.display, root, icon_size, 1)
        XFree(icon_size)

    def setClassHint(self, Window xwindow, wmclass: str, wmname: str) -> None:
        self.context_check("setClassHint")
        cdef XClassHint *classhints = XAllocClassHint()
        assert classhints!=NULL
        res_class = wmclass.encode("latin1")
        res_name = wmname.encode("latin1")
        classhints.res_class = res_class
        classhints.res_name = res_name
        XSetClassHint(self.display, xwindow, classhints)
        XFree(classhints)

    def getClassHint(self, Window xwindow) -> Optional[Tuple[str,str]]:
        self.context_check("getClassHint")
        cdef XClassHint *classhints = XAllocClassHint()
        assert classhints!=NULL
        cdef Status s = XGetClassHint(self.display, xwindow, classhints)
        if not s:
            return None
        _name = ""
        _class = ""
        if classhints.res_name!=NULL:
            _name = classhints.res_name[:]
        if classhints.res_class!=NULL:
            _class = classhints.res_class[:]
        XFree(classhints)
        log("XGetClassHint(%#x) classhints: %s, %s", xwindow, _name, _class)
        return (_name, _class)

    def getGeometry(self, Drawable d) -> Tuple[int, int, int, int, int, int]:
        self.context_check("getGeometry")
        cdef Window root_return
        cdef int x, y                                           #@pydev dupe
        cdef unsigned int width, height, border_width, depth    #@pydev dupe
        if not XGetGeometry(self.display, d, &root_return,
                        &x, &y, &width, &height, &border_width, &depth):
            return None
        return x, y, width, height, border_width, depth

    def getParent(self, Window xwindow) -> XID:
        self.context_check("getParent")
        cdef Window root = 0, parent = 0
        cdef Window *children = NULL
        cdef unsigned int nchildren = 0
        if not XQueryTree(self.display, xwindow, &root, &parent, &children, &nchildren):
            return 0
        if nchildren > 0 and children != NULL:
            XFree(children)
        if parent == XNone:
            return 0
        return int(parent)

    def get_children(self, Window xwindow) -> List[Window]:
        cdef Window root = 0, parent = 0
        cdef Window * children = <Window *> 0
        cdef unsigned int i, nchildren = 0
        if not XQueryTree(self.display, xwindow, &root, &parent, &children, &nchildren):
            return []
        cdef object pychildren = []
        for i in range(nchildren):
            if children[i]>0:
                pychildren.append(children[i])
        # Apparently XQueryTree sometimes returns garbage in the 'children'
        # pointer when 'nchildren' is 0, which then leads to a segfault when we
        # try to XFree the non-NULL garbage.
        if nchildren > 0 and children != NULL:
            XFree(children)
        return pychildren

    def getSizeHints(self, Window xwindow) -> Dict[str,Any]:
        self.context_check("getSizeHints")
        cdef XSizeHints *size_hints = XAllocSizeHints()
        cdef long supplied_return   #ignored!
        assert size_hints!=NULL
        if not XGetWMNormalHints(self.display, xwindow, size_hints, &supplied_return):
            XFree(size_hints)
            return {}
        hints = {}
        if (size_hints.flags & USPosition) or (size_hints.flags & PPosition):
            hints["position"] = size_hints.x, size_hints.y
        if (size_hints.flags & USSize) or (size_hints.flags & PSize):
            hints["size"] = size_hints.width, size_hints.height
        if size_hints.flags & PMinSize:
            hints["min_size"] = size_hints.min_width, size_hints.min_height
        if size_hints.flags & PMaxSize:
            hints["max_size"] = size_hints.max_width, size_hints.max_height
        if size_hints.flags & PBaseSize:
            hints["base_size"] = size_hints.base_width, size_hints.base_height
        if size_hints.flags & PResizeInc:
            hints["resize_inc"] = size_hints.width_inc, size_hints.height_inc
        if size_hints.flags & PAspect:
            try:
                hints["min_aspect_ratio"] = size_hints.min_aspect.x, size_hints.min_aspect.y
                hints["max_aspect_ratio"] = size_hints.max_aspect.x, size_hints.max_aspect.y
            except ZeroDivisionError:
                pass
        if size_hints.flags & PWinGravity:
            hints["win_gravity"] = size_hints.win_gravity
        XFree(size_hints)
        return hints

    def setSizeHints(self, Window xwindow, hints : Dict[str,Any]) -> None:
        self.context_check("setSizeHints")
        cdef XSizeHints *size_hints = XAllocSizeHints()
        assert size_hints!=NULL
        position = hints.get("position")
        if position is not None:
            size_hints.flags |= USPosition | PPosition
            size_hints.x, size_hints.y = position
        size = hints.get("size")
        if size is not None:
            size_hints.flags |= USSize | PSize
            size_hints.width, size_hints.height = size
        min_size = hints.get("min_size")
        if min_size is not None:
            size_hints.flags |= PMinSize
            size_hints.min_width, size_hints.min_height = min_size
        max_size = hints.get("max_size")
        if max_size is not None:
            size_hints.flags |= PMaxSize
            size_hints.max_width, size_hints.max_height = max_size
        base_size = hints.get("base_size")
        if base_size is not None:
            size_hints.flags |= PBaseSize
            size_hints.base_width, size_hints.base_height = base_size
        resize_inc = hints.get("resize_inc")
        if resize_inc is not None:
            size_hints.flags |= PResizeInc
            size_hints.width_inc, size_hints.height_inc = resize_inc
        aspect_ratio = hints.get("aspect-ratio")
        if aspect_ratio is not None:
            size_hints.flags |= PAspect
            size_hints.min_aspect.x, size_hints.min_aspect.y = aspect_ratio
        win_gravity = hints.get("win_gravity")
        if win_gravity is not None:
            size_hints.flags |= PWinGravity
            size_hints.win_gravity = win_gravity
        XSetWMNormalHints(self.display, xwindow, size_hints)
        XFree(size_hints)

    def getWMHints(self, Window xwindow) -> Dict[str,Any]:
        self.context_check("getWMHints")
        cdef XWMHints *wm_hints = XGetWMHints(self.display, xwindow)
        if wm_hints==NULL:
            return None
        hints = {}
        if wm_hints.flags & InputHint:
            hints["input"] = wm_hints.input
        if wm_hints.flags & StateHint:
            hints["initial_state"] = wm_hints.initial_state
        if wm_hints.flags & IconPixmapHint:
            hints["icon_pixmap"] = wm_hints.icon_pixmap
        if wm_hints.flags & IconWindowHint:
            hints["icon_window"] = wm_hints.icon_window
        if wm_hints.flags & IconPositionHint:
            hints["icon_position"] = wm_hints.icon_x, wm_hints.icon_y
        if wm_hints.flags & IconMaskHint:
            hints["icon_mask"] = wm_hints.icon_mask
        if wm_hints.flags & WindowGroupHint:
            hints["window_group"] = wm_hints.window_group
        if wm_hints.flags & XUrgencyHint:
            hints["urgency"] = True
        XFree(wm_hints)
        return hints

    def XGetWMProtocols(self, Window xwindow) -> List[str]:
        self.context_check("XGetWMProtocols")
        cdef Atom *protocols_return
        cdef int count_return
        cdef int i = 0
        protocols = []
        if XGetWMProtocols(self.display, xwindow, &protocols_return, &count_return):
            while i<count_return:
                protocols.append(self.get_atom_name(protocols_return[i]))
                i += 1
        return protocols

    def pointer_grab(self, Window xwindow) -> bool:
        self.context_check("pointer_grab")
        cdef Cursor cursor = 0
        cdef unsigned int event_mask = PointerMotionMask
        cdef int r = XGrabPointer(self.display, xwindow, True,
                                  event_mask, GrabModeAsync, GrabModeAsync,
                                  xwindow, cursor, CurrentTime)
        return r == 0

    def get_server_time(self, Window xwindow) -> long:
        cdef unsigned char c = b"a"
        cdef Atom timestamp_prop = self.str_to_atom("XPRA_TIMESTAMP_PROP")
        XChangeProperty(self.display, xwindow, timestamp_prop,
                   timestamp_prop,
                   8, PropModeReplace, &c, 1)
        cdef XEvent xevent
        cdef xifevent_timestamp et
        et.window = xwindow
        et.atom = timestamp_prop
        XIfEvent(self.display, &xevent, <void*> &timestamp_predicate, <XPointer> &et)
        return xevent.xproperty.time


ctypedef struct xifevent_timestamp:
    Window window
    Atom atom


cdef Bool timestamp_predicate(Display *display, XEvent  *xevent, XPointer arg) nogil:
    cdef xifevent_timestamp *et = <xifevent_timestamp*> arg
    cdef Window xwindow = <Window> arg
    if xevent.type!=PropertyNotify:
        return False
    return xevent.xproperty.window==et.window and xevent.xproperty.atom==et.atom


cdef X11WindowBindingsInstance singleton = None


def X11WindowBindings() -> X11WindowBindingsInstance:
    global singleton
    if singleton is None:
        singleton = X11WindowBindingsInstance()
    return singleton



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


def test_dbus(display, wid=1, flag=True):
    import dbus
    bus = dbus.SessionBus()
    server = 'org.xpra.Server%i' % display
    service = bus.get_object(server, "/org/xpra/Server")
    SetVideoRegionEnabled = service.get_dbus_method('SetVideoRegionEnabled', 'org.xpra.Server')
    print("calling %s.SetVideoRegionEnabled(%i, %s)" % (server, wid, flag))
    SetVideoRegionEnabled(wid, flag)


def main():
    import sys
    if len(sys.argv)!=4:
        print("usage: %s DISPLAY WID True|False" % sys.argv[0])
        return
    DISPLAY = int(sys.argv[1])
    wid = int(sys.argv[2])
    from xpra.scripts.config import str_to_bool
    flag = str_to_bool(sys.argv[3])
    test_dbus(DISPLAY, wid, flag)


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from collections import deque
from typing import Deque
from collections.abc import Callable, Sequence

from xpra.common import SocketState
from xpra.platform.info import get_username
from xpra.platform.dotxpra import DotXpra
from xpra.platform.paths import get_socket_dirs
from xpra.util.objects import typedict
from xpra.util.str_fn import std, obsc, hexstr
from xpra.util.env import envint
from xpra.scripts.config import TRUE_OPTIONS
from xpra.net.digest import get_salt, choose_digest, verify_digest, gendigest
from xpra.os_util import POSIX, is_admin
from xpra.log import Logger

log = Logger("auth")

USED_SALT_CACHE_SIZE = envint("XPRA_USED_SALT_CACHE_SIZE", 1024 * 1024)
DEFAULT_UID = os.environ.get("XPRA_AUTHENTICATION_DEFAULT_UID", "nobody")
DEFAULT_GID = os.environ.get("XPRA_AUTHENTICATION_DEFAULT_GID", "nobody")

# uid, gid, displays, env_options, session_options
SessionData = tuple[int, int, list[str], dict[str, str], dict[str, str]]


def xor(s1, s2) -> bytes:
    return b"".join(b"%c" % (a ^ b) for a, b in zip(s1, s2))


class SysAuthenticatorBase:
    USED_SALT: Deque[bytes] = deque(maxlen=USED_SALT_CACHE_SIZE)
    DEFAULT_PROMPT = "password for user '{username}'"
    CLIENT_USERNAME = False
    VERIFY_USERNAME = is_admin()

    def __init__(self, **kwargs):
        self.username = kwargs.get("username", get_username())
        remote_props = typedict(kwargs.get("remote", {}))
        remote_username = remote_props.strget("username", "")
        verify_username = str(kwargs.get("verify-username", self.VERIFY_USERNAME)).lower() in TRUE_OPTIONS
        client_username = str(kwargs.get("client-username", self.CLIENT_USERNAME)).lower() in TRUE_OPTIONS
        log(f"{self!r}: {verify_username=}, {client_username=}")
        if client_username:
            # allow the client to specify the username to authenticate with:
            self.username = kwargs.get("remote", {}).get("username", self.username)
            log(f"username={self.username!r}")
        elif verify_username:
            self.verify_username(remote_username)
        self.salt = None
        self.digest = None
        self.salt_digest = None
        prompt_attr = {"username": std(self.username)}
        self.prompt = kwargs.pop("prompt", self.DEFAULT_PROMPT).format(**prompt_attr)
        self.socket_dirs = kwargs.pop("socket-dirs", get_socket_dirs())
        self.challenge_sent = False
        self.passed = False
        self.password_used = None
        self.authenticate_check: Callable[[typedict], bool] = self.default_authenticate_check
        # we can't warn about unused options
        # because the options are shared with other socket options (nodelay, cork, etc)
        # unused = dict((k,v) for k,v in kwargs.items() if k not in ("connection", "exec_cwd", "username"))
        # if unused:
        #    log.warn("Warning: unused keyword arguments for %s authentication:", self)
        #    log.warn(" %s", unused)
        log("auth prompt=%s, socket_dirs=%s", self.prompt, self.socket_dirs)

    def verify_username(self, remote_username: str) -> None:
        if remote_username != self.username:
            log(f"verifying username={self.username!r} vs remote={remote_username!r}")
            raise ValueError(f"invalid username {remote_username!r}")

    def get_uid(self) -> int:
        raise NotImplementedError()

    def get_gid(self) -> int:
        raise NotImplementedError()

    def requires_challenge(self) -> bool:
        return True

    def req_xor(self, digests):
        self.req_challenge(digests, "xor")

    def req_challenge(self, digests, required="xor") -> None:
        if required not in digests:
            raise ValueError(f"{self!r} authenticator requires the {required!r} digest")

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        if self.salt is not None:
            log.error("Error: authentication challenge already sent!")
            return b"", ""
        return self.do_get_challenge(digests)

    def do_get_challenge(self, digests) -> tuple[bytes, str]:
        self.salt = get_salt()
        self.digest = choose_digest(digests)
        self.challenge_sent = True
        return self.salt, self.digest

    def get_passwords(self) -> Sequence[str]:
        """ this default implementation just returns a tuple
            with the password from `get_password` if there is one """
        p = self.get_password()  # pylint: disable=assignment-from-none
        if p:
            if not isinstance(p, str):
                raise ValueError(f"password value is not a string: {type(p)}")
            return (p,)
        return ()

    def get_password(self) -> str:
        return ""

    def check(self, value: bytes) -> bool:
        try:
            password = value.decode("utf8")
        except UnicodeDecodeError:
            password = value.decode("latin1")
        return self.check_password(password)

    def check_password(self, password: str) -> bool:
        return password and self.get_password() == password

    def authenticate(self, caps: typedict) -> bool:
        r = self.do_authenticate(caps)
        if r:
            self.passed = True
            log("authentication challenge passed for %s", self)
        return r

    def validate_caps(self, caps: typedict) -> bool:
        if self.passed:
            log("invalid state: challenge has already been passed")
            return False
        if not caps:
            log("invalid state: no capabilities")
            return False
        if not self.challenge_sent:
            log("invalid state: challenge has not been sent yet!")
            return False
        challenge_response = caps.bytesget("challenge_response")
        # challenge has been sent already for this module
        if not challenge_response:
            log("invalid state: challenge already sent but no response found!")
            return False
        return True

    def do_authenticate(self, caps: typedict) -> bool:
        if not self.validate_caps(caps):
            return False
        return self.authenticate_check(caps)

    def choose_salt_digest(self, digest_modes) -> str:
        self.salt_digest = choose_digest(digest_modes)
        return self.salt_digest

    def get_response_salt(self, client_salt=None) -> bytes:
        server_salt = self.salt
        # make sure it does not get re-used:
        self.salt = None
        if client_salt is None:
            return server_salt
        salt = gendigest(self.salt_digest, client_salt, server_salt)
        if salt in SysAuthenticator.USED_SALT:
            raise RuntimeError("danger: an attempt was made to re-use the same computed salt")
        log("combined salt using %r: (%s, %s)=%r", self.salt_digest, hexstr(server_salt), hexstr(client_salt), salt)
        log("salt=%r", salt)
        SysAuthenticator.USED_SALT.append(salt)
        return salt

    def unxor_response(self, caps: typedict) -> bytes:
        challenge_response = caps.bytesget("challenge_response")
        client_salt = caps.strget("challenge_client_salt")
        if self.salt is None:
            log.error("Error: illegal challenge response received - salt cleared or unset")
            return b""
        salt = self.get_response_salt(client_salt)
        value = gendigest("xor", challenge_response, salt)
        log(f"unxor_response(..) challenge-response={obsc(repr(challenge_response))}")
        log(f"client-salt={client_salt!r}, response salt={salt!r}")
        return value

    def default_authenticate_check(self, caps: typedict) -> bool:
        value: bytes = self.unxor_response(caps)
        # warning: enabling logging here would log the actual system password!
        # log.info("authenticate(%s, %s) password=%s (%s)",
        #    hexstr(challenge_response), hexstr(client_salt), password, hexstr(password))
        # verify login:
        try:
            ret = self.check(value)
            log(f"authenticate_check(..)={ret}")
        except Exception as e:
            log("check(..)", exc_info=True)
            log.error(f"Error: {self} authentication check failed:")
            log.estr(e)
            return False
        return ret

    def authenticate_hmac(self, caps: typedict) -> bool:
        challenge_response = caps.bytesget("challenge_response")
        client_salt = caps.strget("challenge_client_salt")
        log("sys_auth_base.authenticate_hmac(%r, %r)", challenge_response, client_salt)
        if not self.salt:
            log.error("Error: illegal challenge response received - salt cleared or unset")
            return False
        salt = self.get_response_salt(client_salt)
        passwords = self.get_passwords()
        if not passwords:
            log.warn(f"Warning: {self} authentication failed")
            log.warn(f" no password defined for {self.username!r}")
            return False
        log(f"found {len(passwords)} passwords using {self!r}")
        for x in passwords:
            if verify_digest(self.digest, x, salt, challenge_response):
                self.password_used = x
                return True
        log.warn(f"Warning: {self.digest} challenge for {self.username!r} does not match")
        if len(passwords) > 1:
            log.warn(f" checked {len(passwords)} passwords")
        return False

    def get_sessions(self) -> SessionData:
        uid = self.get_uid()
        gid = self.get_gid()
        log(f"{self}.get_sessions() uid={uid}, gid={gid}")
        try:
            sockdir = DotXpra(None, self.socket_dirs, actual_username=self.username, uid=uid, gid=gid)
            results = sockdir.sockets(check_uid=uid)
            displays = []
            for state, display in results:
                if state == SocketState.LIVE and display not in displays:
                    displays.append(display)
            log(f"sockdir={sockdir}, results={results}, displays={displays}")
        except Exception as e:
            log("get_sessions()", exc_info=True)
            log.error(f"Error: cannot get the list of sessions for {self.username!r}:")
            log.estr(e)
            displays = []
        v: SessionData = (uid, gid, displays, {}, {})
        log(f"{self}.get_sessions()={v}")
        return v


class SysAuthenticator(SysAuthenticatorBase):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.pw = None
        if POSIX:
            try:
                import pwd  # pylint: disable=import-outside-toplevel
                self.pw = pwd.getpwnam(self.username)
            except (OSError, KeyError):
                log(f"cannot load password database entry for {self.username!r}", exc_info=True)

    def get_uid(self) -> int:
        if self.pw is None:
            raise RuntimeError(f"username {self.username!r} not found")
        return self.pw.pw_uid

    def get_gid(self) -> int:
        if self.pw is None:
            raise RuntimeError(f"username {self.username!r} not found")
        return self.pw.pw_gid



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


import unittest

from xpra.util.objects import AdHocStruct
from xpra.codecs.image import ImageWrapper
from xpra.common import noop
from xpra.server.rfb.source import RFBSource


class TestRFB(unittest.TestCase):

    def test_rfb_source(self):
        # fake protocol:
        p = AdHocStruct()
        p.send = noop
        p.queue_size = lambda : 1
        # fake window:
        window = AdHocStruct()

        def get_image(x, y, w, h):
            stride = (w+8)*4
            pixels = b"0"*stride*h
            return ImageWrapper(x, y, w, h, pixels, "BGRX", 24, stride, 4)
        window.get_image = get_image
        window.acknowledge_changes = noop
        for protocol in (p, None):
            s = RFBSource(protocol, True)
            assert s.get_info()
            s.get_window_info(())
            s.ping()
            s.keys_changed()
            s.set_default_keymap()
            s.send_cursor()
            s.send_server_event("fake-event")
            s.update_mouse()
            s.damage(1, window, 0, 0, 1024, 768, {"polling" : protocol is None})
            s.damage(1, window, 0, 0, 2, 2, {"polling" : protocol is None})
            s.send_clipboard("foo")
            s.bell()
            assert not s.is_closed()
            s.close()
            # noop:
            s.damage(1, window, 0, 0, 2, 2, {"polling" : protocol is None})
            assert s.is_closed()


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.notifications.notifier_base import NotifierBase, NID

Notify = gi_import("Notify")


class GINotifier(NotifierBase):

    def show_notify(self, dbus_id, tray, nid: NID,
                    app_name: str, replaces_nid: NID, app_icon,
                    summary: str, body: str, actions, hints, timeout: int, icon) -> None:
        if not self.dbus_check(dbus_id):
            return
        icon_string = self.get_icon_string(nid, app_icon, icon)
        Notify.init(app_name or "Xpra")
        n = Notify.Notification.new(summary=summary, body=body, icon=icon_string)

        def closed(*_args) -> None:
            self.clean_notification(nid)

        n.connect("closed", closed)
        n.show()

    def close_notify(self, nid: NID) -> None:
        self.clean_notification(nid)

    def cleanup(self) -> None:
        Notify.uninit()
        super().cleanup()



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from ctypes import sizeof, byref, c_char
from ctypes.wintypes import MAX_PATH

from xpra.util.str_fn import strtobytes
from xpra.platform.win32 import constants as win32con
from xpra.log import Logger

# user32:
from xpra.platform.win32.common import (
    GetDC, CreateCompatibleDC, CreateCompatibleBitmap, SelectObject, DeleteObject,
    ReleaseDC, DeleteDC, DrawIconEx, GetBitmapBits,
    GetIconInfo, ICONINFO, Bitmap, GetIconInfoExW, ICONINFOEXW,
    GetObjectA,
)

log = Logger("cursor")

UINT_MAX = 2 ** 32 - 1


def get_cursor_data(hCursor) -> list | None:
    # w, h = get_fixed_cursor_size()
    if not hCursor:
        return None
    dc = None
    memdc = None
    bitmap = None
    old_handle = None
    try:
        ii = ICONINFO()
        ii.cbSize = sizeof(ICONINFO)
        if not GetIconInfo(hCursor, byref(ii)):
            raise OSError()  # @UndefinedVariable
        x = ii.xHotspot
        y = ii.yHotspot
        log("get_cursor_data(%#x) hotspot at %ix%i, hbmColor=%#x, hbmMask=%#x",
            hCursor, x, y, ii.hbmColor or 0, ii.hbmMask or 0)
        if not ii.hbmColor:
            # FIXME: we don't handle black and white cursors
            return None
        iie = ICONINFOEXW()
        iie.cbSize = sizeof(ICONINFOEXW)
        if not GetIconInfoExW(hCursor, byref(iie)):
            raise OSError()  # @UndefinedVariable
        name = iie.szResName[:MAX_PATH]
        log("wResID=%#x, sxModName=%s, szResName=%s", iie.wResID, iie.sxModName[:MAX_PATH], name)
        bm = Bitmap()
        if not GetObjectA(ii.hbmColor, sizeof(Bitmap), byref(bm)):
            raise OSError()  # @UndefinedVariable
        log("cursor bitmap: type=%i, width=%i, height=%i, width bytes=%i, planes=%i, bits pixel=%i, bits=%#x",
            bm.bmType, bm.bmWidth, bm.bmHeight, bm.bmWidthBytes, bm.bmPlanes, bm.bmBitsPixel, bm.bmBits or 0)
        w = bm.bmWidth
        h = bm.bmHeight
        dc = GetDC(None)
        assert dc, "failed to get a drawing context"
        memdc = CreateCompatibleDC(dc)
        assert memdc, "failed to get a compatible drawing context from %s" % dc
        bitmap = CreateCompatibleBitmap(dc, w, h)
        assert bitmap, "failed to get a compatible bitmap from %s" % dc
        old_handle = SelectObject(memdc, bitmap)

        # check if icon is animated:
        if not DrawIconEx(memdc, 0, 0, hCursor, w, h, UINT_MAX, 0, 0):
            log("cursor is animated!")

        # if not DrawIcon(memdc, 0, 0, hCursor):
        if not DrawIconEx(memdc, 0, 0, hCursor, w, h, 0, 0, win32con.DI_NORMAL):
            raise OSError()  # @UndefinedVariable

        buf_size = bm.bmWidthBytes * h
        buftype = c_char * buf_size
        buf = buftype()
        buf.value = b""
        r = GetBitmapBits(bitmap, buf_size, byref(buf))
        log("get_cursor_data(%#x) GetBitmapBits(%#x, %#x, %#x)=%i", hCursor, bitmap, buf_size, byref(buf), r)
        if not r:
            log.error("Error: failed to copy screen bitmap data")
            return None
        elif r != buf_size:
            log.warn("Warning: invalid cursor buffer size, got %i bytes but expected %i", r, buf_size)
            return None
        else:
            # 32-bit data:
            pixels = bytearray(strtobytes(buf.raw))
            has_alpha = False
            has_pixels = False
            for i in range(len(pixels) // 4):
                has_pixels = has_pixels or pixels[i * 4] != 0 or pixels[i * 4 + 1] != 0 or pixels[i * 4 + 2] != 0
                has_alpha = has_alpha or pixels[i * 4 + 3] != 0
                if has_pixels and has_alpha:
                    break
            if has_pixels and not has_alpha:
                # generate missing alpha - don't ask me why
                for i in range(len(pixels) // 4):
                    if pixels[i * 4] != 0 or pixels[i * 4 + 1] != 0 or pixels[i * 4 + 2] != 0:
                        pixels[i * 4 + 3] = 0xff
        return [0, 0, w, h, x, y, hCursor, bytes(pixels), strtobytes(name)]
    except Exception as e:
        log("get_cursor_data(%#x)", hCursor, exc_info=True)
        log.error("Error: failed to grab cursor:")
        log.estr(e)
        return None
    finally:
        if old_handle:
            SelectObject(memdc, old_handle)
        if bitmap:
            DeleteObject(bitmap)
        if memdc:
            DeleteDC(memdc)
        if dc:
            ReleaseDC(None, dc)
